import json
import requests
from cli.utils import download_url, fetch_auth, parse_filter
from zpy.files import to_pathlib_path


@fetch_auth
def create_dataset(name, file_path, project, url, auth_headers):
    """create dataset

    Create dataset on ZumoLabs backend which groups files.

    Args:
        name (str): name of dataset
        file_path (list): zip file to upload
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    endpoint = f"{url}/api/v1/datasets/"
    data = {"name": name, "project": project}
    if file_path:
        r = requests.post(
            endpoint,
            data=data,
            files={"file": open(file_path, "rb")},
            headers=auth_headers,
        )
    else:
        r = requests.post(
            endpoint,
            data=data,
            headers=auth_headers,
        )
    if r.status_code != 201:
        r.raise_for_status()
    return json.loads(r.text)


@fetch_auth
def generate_dataset(dataset_name, sim_name, count, config, project, url, auth_headers):
    """generate dataset

    Generate files for a dataset on ZumoLabs backend which will launch
    a generation job with specified params.

    Args:
        dataset_name (str): name of dataset
        sim_name (str): name of sim to generate from
        count (int): number of times to run the sim
        config (dict): configration of sim for this dataset
        project (str): project uuid
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    from cli.sims import fetch_sim

    fetch_sim(sim_name)
    dataset = create_dataset(dataset_name, None, project)
    endpoint = f"{url}/api/v1/datasets/{dataset['id']}/generate/"
    data = {
        "sim": sim_name,
        "amount": count,
        "config": json.dumps(config),
        "project": project,
    }
    r = requests.post(
        endpoint,
        data=data,
        headers=auth_headers,
    )
    if r.status_code != 200:
        r.raise_for_status()


@fetch_auth
def download_dataset(name, path, url, auth_headers):
    """download dataset

    Download dataset object from S3 through ZumoLabs backend.

    Args:
        name (str): name of dataset to download
        path (str): output directory
        url (str): backend endpoint
        auth_headers: authentication for backend

    Returns:
        str: output file path
    """
    dataset = fetch_dataset(name)
    endpoint = f"{url}/api/v1/datasets/{dataset['id']}/download/"
    r = requests.get(endpoint, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    dataset = json.loads(r.text)
    name_slug = f"{name.replace(' ', '_')}.zip"
    output_path = to_pathlib_path(path) / name_slug
    download_url(dataset["redirect_link"], output_path)
    return output_path


@fetch_auth
def fetch_datasets(filters, url, auth_headers):
    """fetch datasets

    Fetch dataset names from backend. This is done through tags.

    Args:
        filters (dict): query param filters for API call
    Returns:
        list: paginated sorted datasets for all types
    """
    endpoint = f"{url}/api/v1/datasets/"
    r = requests.get(endpoint, headers=auth_headers, params=filters)
    if r.status_code != 200:
        r.raise_for_status()
    return json.loads(r.text)["results"]


@fetch_auth
def fetch_dataset(name, url, auth_headers):
    """fetch dataset

    Fetch info on a dataset by name from backend.

    Args:
        name (str): name of dataset
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    endpoint = f"{url}/api/v1/datasets/"
    r = requests.get(endpoint, params={"name": name}, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    response = json.loads(r.text)
    if response["count"] != 1:
        raise NameError(f"found {response['count']} datasets for name {name}")
    return response["results"][0]


@fetch_auth
def filter_datasets(dfilter, project, url, auth_headers):
    """filter datasets

    Filter dataset objects on ZumoLabs backend by given dfilter.
    Parse dfilter using parse_filter.

    Args:
        dfilter (str): filter query for datasets
        project (str): project uuid
        url (str): backend endpoint
        auth_headers: authentication for backend

    Return:
        dict: filtered datasets by dfilter {'id': 'name'}
    """
    filtered_datasets = {}
    field, pattern, regex = parse_filter(dfilter)
    endpoint = f"{url}/api/v1/datasets/"
    params = {
        "project": project,
        f"{field}__{pattern}": regex,
    }
    while endpoint is not None:
        r = requests.get(endpoint, headers=auth_headers, params=params)
        if r.status_code != 200:
            r.raise_for_status()
        response = json.loads(r.text)
        for dataset in response["results"]:
            filtered_datasets[dataset["id"]] = dataset["name"]
        endpoint = response["next"]
    return filtered_datasets



"""
    File utilities.
"""
import csv
import json
import logging
import os
import random
import re
import shutil
import subprocess
import sys
import tempfile
import zipfile
from pathlib import Path
from pprint import pformat
from typing import Any, Dict, List, Union

log = logging.getLogger(__name__)


"""
Dictionary of filename extensions and prefix/suffixes

These serve as the master search patterns so update and use
these as much as possible to prevent bugs.

Order matters! More specific regexes at the top and
catch-alls at the bottom.

You can test these out at: https://regex101.com/
"""
IMAGE_REGEX = ".*\.(jpeg|jpg|png|bmp)"
FILE_REGEX = {
    # Images
    "instance segmentation image": ".*iseg" + IMAGE_REGEX,
    "class segmentation image": ".*cseg" + IMAGE_REGEX,
    "depth image": ".*depth" + IMAGE_REGEX,
    "normal image": ".*normal" + IMAGE_REGEX,
    "stereo left image": ".*.stereoL" + IMAGE_REGEX,
    "stereo right image": ".*.stereoR" + IMAGE_REGEX,
    "rgb image": ".*rgb" + IMAGE_REGEX,
    "image": IMAGE_REGEX,
    # Annotations
    "zumo annotation": "_annotations.zumo.json",
    "coco annotation": ".*coco.*\.json",
    "annotation": ".*\.(json|xml|yaml|csv)",
}


def dataset_contents(
    path: Union[Path, str],
    filetype_regex: Dict = FILE_REGEX,
) -> Dict:
    """Use regex to search inside a data directory.

    Args:
        path (Union[Path, str]): Directory filepath.
        filetype_regex (Dict, optional): dictionary of {filetype : regex}

    Returns:
        Dict: Contents of directory.
    """
    path = verify_path(path, check_dir=True, make=False)
    contents = {
        "dirs": [],
    }
    for dirpath, _, files in os.walk(path):
        contents["dirs"].append(dirpath)
        for filename in files:
            for name, re_pattern in filetype_regex.items():
                if re.search(re_pattern, filename):
                    if contents.get(name, None) is None:
                        contents[name] = []
                    contents[name].append(os.path.join(dirpath, filename))
                    break
    return contents


def file_is_of_type(
    path: Union[Path, str],
    filetype: str,
) -> bool:
    """Check to if file is of type given.

    Args:
        path (Union[Path, str]): A filesystem path.
        filetype (str): Type of file (see FILE_REGEX dict in zpy.files)

    Returns:
        bool: File is that type.
    """
    if isinstance(path, Path):
        path = str(path)
    assert (
        FILE_REGEX.get(filetype, None) is not None
    ), f"{filetype} must be in {FILE_REGEX.keys()}"
    if re.search(FILE_REGEX[filetype], path):
        return True
    return False


def make_rgb_image_name(id: int, extension: str = ".png") -> str:
    """Creates a RGB image name given an integer id.

    Args:
        id (int): Integer id used in name creation.
        extension (str, optional): Extension for image. Defaults to '.png'.

    Returns:
        str: Image name.
    """
    return "image.%06d.rgb" % id + extension


def make_cseg_image_name(id: int, extension: str = ".png") -> str:
    """Return category (class) segmentation image name from integer id.

    Args:
        id (int): Integer id used in name creation.
        extension (str, optional): Extension for image. Defaults to '.png'.

    Returns:
        str: Image name.
    """
    return "image.%06d.cseg" % id + extension


def make_iseg_image_name(id: int, extension: str = ".png") -> str:
    """Return instance segmentation image name from integer id.

    Args:
        id (int): Integer id used in name creation.
        extension (str, optional): Extension for image. Defaults to '.png'.

    Returns:
        str: Image name.
    """
    return "image.%06d.iseg" % id + extension


def make_depth_image_name(id: int, extension: str = ".png") -> str:
    """Return depth image name from integer id.

    Args:
        id (int): Integer id used in name creation.
        extension (str, optional): Extension for image. Defaults to '.png'.

    Returns:
        str: Image name.
    """
    return "image.%06d.depth" % id + extension


def make_custom_image_name(id: int, name: str, extension: str = ".png") -> str:
    """Creates a custom image name given integer id and name.

    Args:
        id (int): Integer id used in name creation.
        name (str): Custom string which will be appended to the end of the image name.
        extension (str, optional): Extension for image. Defaults to '.png'.

    Returns:
        str: Image name.
    """
    return "image.%06d.%s" % (id, name) + extension


def id_from_image_name(image_name: str) -> int:
    """Extract integer id from image name.

    Args:
        image_name (str): Name of image to extract integer from.

    Returns:
        int: Integer id.
    """
    return int("".join([s for s in image_name if s.isdigit()]))


def replace_id_in_image_name(image_name: str, new_id: int) -> str:
    """Replace the integer id in an image name.

    Args:
        image_name (str): Name of the image.
        new_id (int): New id to replace old id with.

    Returns:
        str: New image name.
    """
    # HACK: This will break for image names without 8-digit indices
    return "image.%06d" % new_id + image_name[12:]


def add_to_path(path: Union[Path, str], name: str) -> Path:
    """Add string descriptor to path: foo.txt -> foo.more_foo.txt

    Args:
        path (Union[Path, str]): A filesystem path.
        name (str): Name to append to file name.

    Returns:
        Path: New path.
    """
    path = to_pathlib_path(path)
    underscore_filename = path.stem + "_" + name + path.suffix
    return path.parent / Path(underscore_filename)


def to_pathlib_path(path: Union[Path, str]) -> Path:
    """Convert string path to pathlib.Path if needed.

    Args:
        path (Union[Path, str]): A filesystem path.

    Returns:
        Path: Path in pathlib.Path format.
    """
    if not isinstance(path, Path):
        path = Path(os.path.expandvars(path)).resolve()
    return path


def default_temp_path() -> Path:
    """Default temporary path agnostic to OS.

    Returns:
        Path: Path to a new output folder in the temp path.
    """
    return Path(tempfile.gettempdir()) / "output"


def clean_dir(
    path: Union[Path, str],
    keep_dir: bool = True,
) -> None:
    """Delete everything at the provided directory.

    Args:
        path (Union[Path, str]): Path to directory.
        keep_dir (bool, optional): Whether to keep (or delete) the directory itself. Defaults to True.
    """
    path = verify_path(path, make=False, check_dir=True)
    if keep_dir:
        # Delete the contents, but keep the directory
        for _path in path.iterdir():
            try:
                if _path.is_file() or _path.is_symlink():
                    _path.unlink()
                elif _path.is_dir():
                    shutil.rmtree(_path)
            except Exception as e:
                log.warning("Failed to delete %s. Reason: %s" % (_path, e))
    else:
        # Delete everything, including the directory itself
        shutil.rmtree(path)


def pretty_print(d: Dict) -> str:
    """Pretty formatted dictionary.

    Args:
        d (Dict): Dictionary to be pretty printed

    Returns:
        str: Dictionary in pretty format.
    """
    return pformat(d, indent=2, width=120)


def verify_path(
    path: Union[Path, str],
    make: bool = False,
    check_dir: bool = False,
) -> Path:
    """Checks to make sure Path exists and optionally creates it.

    Args:
        path (Union[Path, str]): A filesystem path.
        make (bool, optional): Make the path if it does not exist. Defaults to False.
        check_dir (bool, optional): Throw error is path is not a directory. Defaults to False.

    Raises:
        ValueError: Path is not a directory (only if check_dir is set to True)

    Returns:
        Path: The same path.
    """
    path = to_pathlib_path(path)
    if not path.exists():
        log.warning(f"Could not find path at {path}")
        if make:
            log.info(f"Making {path.name} dir at {path}")
            path.mkdir(exist_ok=True, parents=True)
    else:
        log.debug(f"Path found at {path}.")
        if check_dir and not path.is_dir():
            raise ValueError(f"Path at {path} is not a directory.")
    return path


def write_json(
    path: Union[Path, str],
    data: Union[Dict, List],
) -> None:
    """Save data to json file.

    Args:
        path (Union[Path, str]): Path to output json.
        data (Union[Dict, List]): Data to save.

    Raises:
        ValueError: Path is not a json file.
    """
    path = to_pathlib_path(path)
    if not path.suffix == ".json":
        raise ValueError(f"{path} is not a JSON file.")
    log.info(f"Writing JSON to file {path}")
    with path.open("w") as f:
        json.dump(data, f, indent=4)


def read_json(
    path: Union[Path, str],
) -> Union[Dict, List]:
    """Read a json from a path.

    Args:
        path (Union[Path, str]): A filesystem path.

    Raises:
        ValueError: Path is not a json file.

    Returns:
        Union[Dict, List]: Data in the json.
    """
    path = to_pathlib_path(path)
    if not path.suffix == ".json":
        raise ValueError(f"{path} is not a JSON file.")
    log.info(f"Reading JSON file at {path}")
    with path.open() as f:
        data = json.load(f)
    return data


def write_csv(
    path: Union[Path, str], data: List[List[Any]], delimiter: str = ",", **kwargs
) -> None:
    """Write data to csv.

    Pass in additional kwargs to the csv writer.

    Args:
        path (Union[Path, str]): A filesystem path.
        data (List[List[Any]]): Data to save.
        delimiter (str, optional): Delimiter for each row of csv. Defaults to ','.

    Raises:
        ValueError: Path is not a csv or txt file.
    """
    path = to_pathlib_path(path)
    if path.suffix not in [".csv", ".txt"]:
        raise ValueError(f"{path} is not a CSV file.")
    log.info(f"Writing CSV to file {path}")
    with path.open("w") as f:
        writer = csv.writer(f, delimiter=delimiter, **kwargs)
        writer.writerows(data)


def read_csv(path: Union[Path, str], delimiter: str = ",", **kwargs) -> List[List[Any]]:
    """Read csv data from a path.

    Pass in additional kwargs to the csv reader.

    Args:
        path (Union[Path, str]): A filesystem path.
        delimiter (str, optional): Delimiter for each row of csv. Defaults to ','.

    Raises:
        ValueError: Path is not a csv or txt file.

    Returns:
        List[List[Any]]: Data in the csv.
    """
    path = to_pathlib_path(path)
    if path.suffix not in [".csv", ".txt"]:
        raise ValueError(f"{path} is not a CSV file.")
    log.info(f"Reading CSV file at {path}")
    data = []
    with path.open() as f:
        for row in csv.reader(f, delimiter=delimiter, **kwargs):
            data.append(row)
    return data


def pick_random_from_dir(
    dir_path: Union[Path, str],
    suffixes: List[str] = [".txt"],
) -> Path:
    """Pick random file of suffix in a directory.

    Args:
        dir_path (Union[Path, str]): Path to the directory containing files.
        suffixes (List[str], optional): Filter files by these suffixes. Defaults to [].

    Returns:
        Path: Path to randomly chosen file with given suffix.
    """
    _paths = []
    for _path in dir_path.iterdir():
        if _path.is_file() and _path.suffix in suffixes:
            _paths.append(_path)
    _path = random.choice(_paths)
    log.debug(f"Found {len(_paths)} files with suffix {suffixes} at {dir_path}")
    log.info(f"Randomly chose {_path}")
    return _path


def sample(
    things: List,
    sample_size: int = None,
) -> List:
    """Sample N things from a list.

    Args:
        things (List): List of things.
        sample_size (int, optional): Sample size N. Defaults to length of things.

    Returns:
        List: New sample of things.
    """
    random_sample_size = len(things)
    if sample_size is not None:
        random_sample_size = min(sample_size, len(things))
    if random_sample_size == len(things):
        sample_images = things
    else:
        sample_images = random.sample(things, random_sample_size)
    return sample_images


def filecopy(
    src_path: Union[Path, str],
    dst_path: Union[Path, str],
) -> None:
    """Copy file from source (src) to destination (dst).

    Args:
        src_path (Union[Path, str]): Source filesystem path.
        dst_path (Union[Path, str]): Destination filesystem path.
    """
    src_path = verify_path(src_path)
    dst_path = verify_path(dst_path)
    log.debug(f"Copying over file from {src_path} to {dst_path}")
    shutil.copy(src_path, dst_path)


def open_folder_in_explorer(
    path: Union[Path, str],
    make: bool = False,
) -> None:
    """Opens a directory in the fileexplorer of your OS.

    Args:
        path (Union[Path, str]): Filesystem path.
        make (bool, optional): Make directory if it doesn't exist. Defaults to False.
    """
    path = verify_path(path, check_dir=True, make=make)
    if sys.platform.startswith("darwin"):
        subprocess.call(("open", path))
    elif os.name == "nt":
        os.startfile(path)
    elif os.name == "posix":
        subprocess.call(("xdg-open", path))


def remove_files_with_suffix(
    path: Union[Path, str],
    exts: List[str],
) -> None:
    """Remove file in a path with certain extension.

    Args:
        path (Union[Path, str]): Directory path.
        exts (List[str]): List of extensions to remove
    """
    path = verify_path(path, check_dir=True)
    for _path in path.glob("*"):
        if _path.suffix in exts:
            log.info(f"Removing file at {_path}")
            _path.unlink()


def unzip_file(
    zip_path: Union[Path, str],
    out_path: Union[Path, str],
) -> None:
    """Unzip a file to an output path.

    Args:
        zip_path (Union[Path, str]): Path to zip file.
        out_path (Union[Path, str]): Path to output directory.

    Raises:
        ValueError: Path isn't a zip.
    """
    log.info(f"Unzipping {zip_path} to {out_path}...")
    zip_path = verify_path(zip_path)
    out_path = verify_path(out_path, check_dir=True)
    if not zip_path.suffix == ".zip":
        raise ValueError(f"{zip_path} is not a zip file")
    zf = zipfile.ZipFile(str(zip_path))
    zipped_size_mb = round(sum([i.compress_size for i in zf.infolist()]) / 1024 / 1024)
    unzipped_size_mb = round(sum([i.file_size for i in zf.infolist()]) / 1024 / 1024)
    log.info(f"Compressed: {zipped_size_mb}MB, actual: {unzipped_size_mb}MB.")
    zf.extractall(out_path)
    log.info(f"Done extracting to {out_path}.")


def zip_file(
    in_path: Union[Path, str],
    zip_path: Union[Path, str],
) -> None:
    """Zip a directory to a path.

    Args:
        in_path (Union[Path, str]): Path to input directory.
        zip_path (Union[Path, str]): Path to zip file.

    Raises:
        ValueError: Path isn't a zip.
    """
    log.info(f"Zipping {in_path} to {zip_path}...")
    in_path = verify_path(in_path)
    zip_path = verify_path(zip_path)
    if not zip_path.suffix == ".zip":
        raise ValueError(f"{zip_path} is not a zip file")
    shutil.make_archive(
        base_name=zip_path.parent / zip_path.stem, format="zip", root_dir=in_path
    )
    log.info(f"Done zipping to {zip_path}.")
    zf = zipfile.ZipFile(str(zip_path))
    zipped_size_mb = round(sum([i.compress_size for i in zf.infolist()]) / 1024 / 1024)
    unzipped_size_mb = round(sum([i.file_size for i in zf.infolist()]) / 1024 / 1024)
    log.info(f"Compressed: {zipped_size_mb}MB, actual: {unzipped_size_mb}MB.")



import json

import click
import requests
from requests.auth import HTTPBasicAuth
from table_logger import TableLogger

from cli.config import initialize_config, read_config, write_config, add_env, swap_env
from cli.loader import Loader
from cli.utils import parse_args, resolve_sweep, use_project, print_list_as_columns
from zpy.files import read_json, to_pathlib_path

SMALL_WIDTH = 12
MEDIUM_WIDTH = 24
LARGE_WIDTH = 36
UUID_WIDTH = 36
DATETIME_WIDTH = 27


@click.group(context_settings=dict(token_normalize_func=str.lower))
def cli():
    """zpy cli

    Zumo Labs cli which is used to create, get, list, upload objects from
    the Zumo Labs backend (ragnarok).
    """
    initialize_config()


@cli.command("help")
def cli_help():
    """display help

    This will display help in order to provide users with more information
    on how to use this tool.
    """
    click.echo(
        "zpy - ZumoLabs command line interface\n"
        "\n"
        "zpy is a tool used to list, create, upload, download\n"
        "objects from the ZumoLabs backend (ragnarok)\n"
        "\n"
        "app    - https://app.zumolabs.ai\n"
        "github - https://github.com/ZumoLabs/zpy\n"
        "docs   - https://github.com/ZumoLabs/zpy/tree/main/docs/cli"
    )


@cli.command("login")
@click.argument("username", required=True)
@click.password_option(help="The login password.")
def login(username, password):
    """login to ragnarok

    This command will update the zpy config with a token that is fetched
    from the backend using account details.

    Accounts can be created at: app.zumolabs.ai

    Args:
        username (str): developer username
        password (str): developer password
    """
    config = read_config()
    endpoint = f"{config['ENDPOINT']}/auth/login/"
    r = requests.post(endpoint, auth=HTTPBasicAuth(username, password))
    if r.status_code != 200:
        click.secho("Login failed.", err=True, fg="red")
        return
    config["TOKEN"] = r.json()["token"]
    write_config(config)
    click.echo("Login successful!")


@cli.command("config")
def cli_config():
    """display config

    Display current configuration file to developer.
    """
    pretty_config = json.dumps(read_config(), indent=2)
    click.echo(f"Zpy cli configuration:\n{pretty_config}")


@cli.command("version")
def version():
    """version

    Display the zpy cli version.
    """
    import zpy

    click.echo(f"Version: {zpy.__version__}")


# ------- ENV


@cli.group("env")
def env_group():
    """environment configuration.

    Configure the environment for backend calls.
    """
    pass


@env_group.command("set")
@click.argument("env")
def set_env(env):
    """switch target environment

    This command allows zumo labs developers to swap the endpoint that the cli communicates with.

    Args:
        env (str): new environment for endpoint
    """
    config = read_config()
    old_env, old_endpoint = config["ENVIRONMENT"], config["ENDPOINT"]
    swap_env(env)
    config = read_config()
    click.echo("Swapped environment:")
    click.echo(f"  {old_env} -> {config['ENVIRONMENT']}")
    click.echo(f"  {old_endpoint} -> {config['ENDPOINT']}")
    click.echo("zpy login to fetch token")


@env_group.command("add")
@click.argument("env")
@click.argument("endpoint")
def add_environment(env, endpoint):
    """add a new environment

    This command allows you to add an environment to target with backend calls.

    Args:
        env (str): new environment name identifier
        endpoint (str): endpoint for new environment
    """
    click.echo(f"Adding environment:")
    click.echo(f"  ENVIRONMENT: {env}")
    click.echo(f"  ENDPOINT: {endpoint}")
    add_env(env, endpoint)


# ------- DATASET


@cli.group("dataset")
def dataset_group():
    """dataset object.

    Dataset is a collection of files.
    """
    pass


@dataset_group.command("list")
@click.argument("filters", nargs=-1)
@use_project()
def list_datasets(filters, project=None):
    """list datasets

    List datasets from backend with optional FILTERS. Uses PROJECT set via zpy project command when available.
    """
    from cli.datasets import fetch_datasets

    try:
        filters = parse_args(filters)
        if project:
            filters["project"] = project
    except Exception:
        click.secho(f"Failed to parse filters: {filters}", fg="yellow", err=True)
        return

    try:
        with Loader("Fetching datasets..."):
            datasets = fetch_datasets(filters)
        click.echo("Fetched datasets successfully.")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to fetch datasets {e}.", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)
        return

    tbl = TableLogger(
        columns="name,state,files,created_at",
        colwidth={
            "name": LARGE_WIDTH,
            "state": MEDIUM_WIDTH,
            "files": SMALL_WIDTH,
            "created_at": DATETIME_WIDTH,
        },
    )
    for d in datasets:
        tbl(
            d["name"],
            d["state"],
            d["num_files"],
            d["created_at"],
        )


@dataset_group.command("get")
@click.argument("name")
@click.argument(
    "path",
    type=click.Path(exists=True, file_okay=False, writable=True, resolve_path=True),
)
@click.argument("format", default="archive")
def get_dataset(name, path, format):
    """get dataset

    Download dataset of type DTYPE and name NAME to local PATH from backend.

    Args:
        name (str): name of dataset
        path (str): directory to put zipped dataset
        format (str): format for packaging
    """
    from cli.datasets import download_dataset
    from cli.utils import download_url

    try:
        output_path = download_dataset(name, path)
        click.echo(f"Downloaded dataset '{name}' to {output_path}")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to download dataset: {e}", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)
    except NameError as e:
        click.secho(f"Failed to download dataset: {e}", fg="yellow", err=True)


@dataset_group.command("upload")
@click.argument("name")
@click.argument("path", type=click.Path(exists=True, dir_okay=False, resolve_path=True))
@use_project(required=True)
def upload_dataset(name, path, project=None):
    """upload dataset

    Upload dataset located at PATH to PROJECT and call it NAME. Requires PROJECT to be set via `zpy project`.

    Args:
        name (str): name of dataset
        path (str): path to dataset
        project (str): project uuid
    """
    from cli.datasets import create_dataset

    if to_pathlib_path(path).suffix != ".zip":
        click.secho(f"File {path} must be of type zip", fg="red", err=True)
    try:
        with Loader("Uploading dataset..."):
            create_dataset(name, path, project)
        click.secho(f"Uploaded dataset {path} with name '{name}'", fg="green")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to upload dataset: {e}", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)


@dataset_group.command("generate")
@click.argument("name")
@click.argument("sim")
@click.argument("number")
@click.argument("args", nargs=-1)
@use_project(required=True)
def create_dataset(name, sim, number, args, project=None):
    """Create a dataset.

    Create a dataset object called NAME. This will trigger the generation of data from SIM with NUMBER of runs given the input ARGS. Requires PROJECT to be set via `zpy project`.

    Args:
        name (str): name of new dataset
        sim (str): name of sim dataset is built with
        number (str): number of datasets to create
        args (List(str)): configuration of sim for this dataset
        project (str): project uuid
    """
    from cli.datasets import generate_dataset

    try:
        dataset_config = parse_args(args)
    except Exception:
        click.secho(f"Failed to parse args: {args}", fg="yellow", err=True)
        return

    try:
        generate_dataset(name, sim, number, dataset_config, project)
        click.secho(
            f"Generating {number} from sim '{sim}' with config {dataset_config}",
            fg="green",
        )
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to create dataset: {e}", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)
    except NameError as e:
        click.secho(f"Failed to create dataset: {e}", fg="yellow", err=True)


# ------- PROJECT


@cli.group("project")
def project_group():
    """Project group

    Project is a container for the rest of the objects.
    """
    pass


@project_group.command("list")
@click.argument("filters", nargs=-1)
def list_projects(filters):
    """list projects

    List projects from backend with optional FILTERS.
    """
    from cli.projects import fetch_projects

    try:
        filters = parse_args(filters)
    except Exception:
        click.secho(f"Failed to parse filters: {filters}", fg="yellow", err=True)
        return

    try:
        with Loader("Fetching projects..."):
            projects = fetch_projects(filters)
        click.echo("Fetched projects successfully.")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to fetch projects {e}.", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)
        return

    tbl = TableLogger(
        columns="id,name,account,created_at",
        colwidth={
            "id": UUID_WIDTH,
            "name": LARGE_WIDTH,
            "account": UUID_WIDTH,
            "created_at": DATETIME_WIDTH,
        },
    )
    for p in projects:
        tbl(
            p["id"],
            p["name"],
            p["account"],
            p["created_at"],
        )


@project_group.command("create")
@click.argument("account", type=click.UUID)
@click.argument("name")
def create_project(account, name):
    """Create a project under ACCOUNT called NAME.

    See available accounts: zpy list accounts
    """
    from cli.projects import create_project

    try:
        create_project(account, name)
        click.secho(f"Created project '{name}'", fg="green")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to create project: {e}", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)


@project_group.command("set")
@click.argument("project_uuid", type=click.UUID)
def set_project(project_uuid):
    """Set project

    Set global PROJECT uuid.
    """
    config = read_config()
    old_project_uuid = config.get("PROJECT", None)
    config["PROJECT"] = str(project_uuid)
    write_config(config)
    click.echo("Switched project:")
    click.echo(f"  {old_project_uuid} -> {config['PROJECT']}")


@project_group.command("clear")
def clear_project():
    """Clear project

    Clear global PROJECT uuid.
    """
    config = read_config()
    config.pop("PROJECT")
    write_config(config)
    click.echo("Cleared global project namespace.")


# ------- SIM


@cli.group("sim")
def sim_group():
    """Sim object

    Sim is a 3D scene which is used to generate images.
    """
    pass


@sim_group.command("list")
@click.argument("filters", nargs=-1)
@use_project()
def list_sims(filters, project=None):
    """list sims

    List sims from backend with optional FILTERS. Uses PROJECT set via zpy project command when available.
    """
    from cli.sims import fetch_sims

    try:
        filters = parse_args(filters)
        if project:
            filters["project"] = project
    except Exception:
        click.secho(f"Failed to parse filters: {filters}", fg="yellow", err=True)
        return

    try:
        with Loader("Fetching sims..."):
            sims = fetch_sims(filters)
        click.echo("Fetched sims successfully.")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to fetch sims {e}.", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)
        return

    tbl = TableLogger(
        columns="name,state,zpy_version,blender_version,created_at",
        colwidth={
            "name": LARGE_WIDTH,
            "state": MEDIUM_WIDTH,
            "zpy_version": MEDIUM_WIDTH,
            "blender_version": SMALL_WIDTH,
            "created_at": DATETIME_WIDTH,
        },
    )
    for s in sims:
        tbl(
            s["name"],
            s["state"],
            s["zpy_version"],
            s["blender_version"],
            s["created_at"],
        )


@sim_group.command("get")
@click.argument("name")
@click.argument(
    "path",
    type=click.Path(exists=True, file_okay=False, writable=True, resolve_path=True),
)
def get_sim(name, path):
    """get sim

    Download sim with name NAME from backend.

    Args:
        name (str): name of sim
        path (str): directory to put zipped sim
    """
    from cli.sims import download_sim

    try:
        output_path = download_sim(name, path)
        click.echo(f"Downloaded sim '{name}' to {output_path}")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to download sim: {e}", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)
    except NameError as e:
        click.secho(f"Failed to download sim: {e}", fg="yellow", err=True)


@sim_group.command("upload")
@click.argument("name")
@click.argument("path", type=click.Path(exists=True, dir_okay=False, resolve_path=True))
@use_project(required=True)
def upload_sim(name, path, project=None):
    """upload sim

    Upload sim located at PATH to PROJECT and call it NAME. Requires PROJECT to be set via `zpy project`.

    Args:
        name (str): name of sim
        path (str): path to sim
        project (str): project uuid
    """
    from cli.sims import create_sim

    if to_pathlib_path(path).suffix != ".zip":
        click.secho(f"File {path} must be of type zip", fg="red", err=True)
    try:
        with Loader("Uploading sim..."):
            create_sim(name, path, project)
        click.secho(f"Uploaded sim {path} with name '{name}'", fg="green")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to upload sim: {e}", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)


@sim_group.command("logs")
@click.argument("name")
@click.argument(
    "path",
    type=click.Path(exists=True, file_okay=False, writable=True, resolve_path=True),
)
def logs_sim(name, path):
    from cli.logs import fetch_logs

    try:
        fetch_logs("sims", name, path)
        click.echo(f"Downloaded {path}/[info/debug/error].log from '{name}'.")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to fetch logs: {e}", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)
    except NameError as e:
        click.secho(f"Failed to fetch logs: {e}", fg="yellow", err=True)


# ------- ACCOUNT


@cli.group("account")
def account_group():
    """Account object

    Accounts are used to interact with the backend.
    """
    pass


@account_group.command("list")
@click.argument("filters", nargs=-1)
def list_accounts(filters):
    """list accounts

    List accounts from backend with optional FILTERS.
    """
    from cli.accounts import fetch_accounts

    try:
        filters = parse_args(filters)
    except Exception:
        click.secho(f"Failed to parse filters: {filters}", fg="yellow", err=True)
        return

    try:
        with Loader("Fetching accounts..."):
            accounts = fetch_accounts(filters)
        click.echo("Fetched accounts successfully.")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to fetch accounts {e}.", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)
        return

    tbl = TableLogger(
        columns="id,type,email,created_at",
        colwidth={
            "id": UUID_WIDTH,
            "type": LARGE_WIDTH,
            "email": UUID_WIDTH,
            "created_at": DATETIME_WIDTH,
        },
    )
    for p in accounts:
        tbl(
            p["id"],
            p["type"],
            p["email"],
            p["created_at"],
        )


# ------- JOB


@cli.group("job")
def job_group():
    """Job object

    Jobs are used in order to perform operations on a set of datasets.
    """
    pass


@job_group.command("list")
@click.argument("filters", nargs=-1)
@use_project()
def list_jobs(filters, project=None):
    """
    list jobs

    List jobs from backend with optional FILTERS. Uses PROJECT set via `zpy project` command when available.
    """
    from cli.jobs import fetch_jobs

    try:
        filters = parse_args(filters)
        if project:
            filters["project"] = project
    except Exception:
        click.secho(f"Failed to parse filters: {filters}", fg="yellow", err=True)
        return

    try:
        with Loader("Fetching jobs..."):
            jobs = fetch_jobs(filters)
        click.echo("Fetched jobs successfully.")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to fetch jobs {e}.", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)
        return

    tbl = TableLogger(
        columns="state,name,operation,created_at",
        colwidth={
            "state": MEDIUM_WIDTH,
            "name": LARGE_WIDTH,
            "operation": SMALL_WIDTH,
            "created_at": DATETIME_WIDTH,
        },
    )
    for j in jobs:
        tbl(j["state"], j["name"], j["operation"], j["created_at"])


@job_group.command("create")
@click.argument("name")
@click.argument("operation", type=click.Choice(["package", "tvt", "train"]))
@click.option(
    "filters",
    "-f",
    multiple=True,
    help="Key/value pairs separated by spaces. Passed as query params in the API call to filter data sets.",
)
@click.option(
    "configfile",
    "--configfile",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="Path to json file",
)
@click.option(
    "sweepfile",
    "--sweepfile",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="Path to json file",
)
@use_project(required=True)
def create_job(name, operation, filters, configfile, sweepfile, project=None):
    """create job

    Create a job called NAME within PROJECT to perform OPERATION on a group of datasets defined by the FILTERS
    provided by -f. Requires PROJECT set via `zpy project`.
    """
    from cli.datasets import filter_datasets
    from cli.jobs import create_new_job

    filtered_datasets = []
    for dfilter in filters:
        try:
            with Loader(f"Filtering datasets by '{dfilter}'..."):
                datasets = filter_datasets(dfilter, project)

            count = len(datasets)
            click.secho(f"Found {count} matching '{dfilter}'")

            if count == 0:
                continue

            dataset_names = list(datasets.values())
            print_list_as_columns(dataset_names)

            filtered_datasets.extend(datasets.keys())
        except requests.exceptions.HTTPError as e:
            click.secho(f"Failed to filter datasets {e}", fg="red", err=True)

    job_configs = []
    if configfile:
        config = read_json(configfile)
        job_configs.append(config)
        click.echo(f"Parsed config file {configfile} : {config}")
    elif sweepfile:
        sweep_config = read_json(sweepfile)
        try:
            configs = resolve_sweep(sweep_config)
        except Exception as e:
            click.secho(
                f"Failed to resolve sweep file {sweepfile} {e}", fg="yellow", err=True
            )
            return
        job_configs.extend(configs)
        click.echo(f"Parsed sweep file {sweepfile} : {sweep_config}")
    else:
        job_configs.append(dict())

    click.confirm(f"Launch {len(job_configs)} jobs?", abort=True)

    for i, config in enumerate(job_configs):
        job_name = name if i == 0 else f"{name} {i}"
        try:
            create_new_job(job_name, operation, config, filtered_datasets, project)
            click.secho(
                f"Created {operation} job '{job_name}' with config {config}", fg="green"
            )
        except requests.exceptions.HTTPError as e:
            click.secho(f"Failed to create job: {e}", fg="red", err=True)
            if e.response.status_code == 400:
                click.secho(str(e.response.json()), fg="red", err=True)

    click.echo(f"Finished creating {len(job_configs)} jobs with name '{name}'")


@job_group.command("logs")
@click.argument("name")
@click.argument(
    "path",
    type=click.Path(exists=True, file_okay=False, writable=True, resolve_path=True),
)
def logs_job(name, path):
    from cli.logs import fetch_logs

    try:
        fetch_logs("jobs", name, path)
        click.echo(f"Downloaded {path}/[info/debug/error].log from '{name}'.")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to fetch logs: {e}", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)
    except NameError as e:
        click.secho(f"Failed to fetch logs: {e}", fg="yellow", err=True)


# ------- TRANSFORM


@cli.group("transform")
def transform_group():
    """Transform Operations

    Transforms are used on datasets to output a new dataset.
    """
    pass


@transform_group.command("list")
@click.argument("filters", nargs=-1)
@use_project()
def list_transforms(filters, project=None):
    """
    list transforms

    List transforms from backend with optional FILTERS. Also displays available TRANSFORMS. Uses PROJECT set via `zpy project` command when available.
    """
    from cli.transforms import fetch_transforms, available_transforms

    try:
        filters = parse_args(filters)
        if project:
            filters["project"] = project
    except Exception:
        click.secho(f"Failed to parse filters: {filters}", fg="yellow", err=True)
        return

    try:
        click.echo(f"Available transforms: {available_transforms()}")
        with Loader("Fetching transforms..."):
            transforms = fetch_transforms(filters)
        click.echo("Fetched transforms successfully.")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to fetch transforms {e}.", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)
        return

    tbl = TableLogger(
        columns="state,operation,input_dataset,created_at",
        colwidth={
            "state": MEDIUM_WIDTH,
            "operation": SMALL_WIDTH,
            "input_dataset": LARGE_WIDTH,
            "created_at": DATETIME_WIDTH,
        },
    )
    for t in transforms:
        tbl(t["state"], t["operation"], t["input_dataset"], t["created_at"])


@transform_group.command("dataset")
@click.argument("name")
@click.argument("operation")
@click.argument("args", nargs=-1)
@use_project(required=True)
def transform_dataset(name, operation, args, project=None):
    """Transform a dataset.

    Transform a dataset NAME with OPERATION. This will trigger the transformation of this dataset given the input ARGS. Requires PROJECT to be set via `zpy project`.

    Args:
        name (str): name of new dataset
        operation (str): operation to run on dataset
        args (List(str)): configuration of sim for this dataset
        project (str): project uuid
    """
    from cli.transforms import create_transform

    try:
        transform_config = parse_args(args)
    except Exception:
        click.secho(f"Failed to parse args: {args}", fg="yellow", err=True)
        return

    try:
        create_transform(name, operation, transform_config, project)
        click.secho(
            f"Running {operation} on dataset '{name}' with config {transform_config}",
            fg="green",
        )
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to create transform: {e}", fg="red", err=True)
        if e.response.status_code == 400:
            click.secho(str(e.response.json()), fg="red", err=True)
    except NameError as e:
        click.secho(f"Failed to create transform: {e}", fg="yellow", err=True)



from itertools import cycle
from shutil import get_terminal_size
from threading import Thread
from time import sleep


class Loader:
    def __init__(self, desc="Loading...", end="Done!", timeout=0.2):
        """
        A loader-like context manager

        Args:
            desc (str, optional): The loader's description. Defaults to "Loading...".
            timeout (float, optional): Sleep time between prints. Defaults to 0.1.
        """
        self.desc = desc
        self.timeout = timeout

        self._thread = Thread(target=self._animate, daemon=True)
        self.steps = ["⢿", "⣻", "⣽", "⣾", "⣷", "⣯", "⣟", "⡿"]
        self.done = False
        self.end = end

    def start(self):
        self._thread.start()
        return self

    def _animate(self):
        for c in cycle(self.steps):
            if self.done:
                break
            print(f"\r{self.desc} {c}", flush=True, end="")
            sleep(self.timeout)

    def __enter__(self):
        self.start()

    def stop(self):
        self.done = True
        cols = get_terminal_size((80, 20)).columns
        print("\r" + " " * cols, end="", flush=True)
        print(f"\r{self.end}", flush=True)

    def __exit__(self, exc_type, exc_value, tb):
        # handle exceptions with those variables ^
        self.stop()



"""
    Base class for new annotation formats. Any new annotation format must be
    able to take a saver object and output annotations.
"""
import logging
from pathlib import Path
from typing import Union

import zpy

log = logging.getLogger(__name__)


class Output:
    """Outputs a Saver object to various annotation file formats."""

    def __init__(
        self,
        saver: zpy.saver.Saver = None,
        annotation_filename: Union[Path, str] = "annotations.txt",
    ):
        """Create an Output object from a Saver object.

        Args:
            saver (zpy.saver.Saver, optional): Saver contains annotation and category information.
            annotation_filename (Union[Path, str], optional): Default name for annotation files.
        """
        if saver is not None:
            self.saver = saver
            # Try and deduce the annotation path from the saver object
            if self.saver.annotation_path is not None:
                self.annotation_path = self.saver.annotation_path
            elif self.saver.output_dir is not None:
                self.annotation_path = self.saver.output_dir / annotation_filename
            else:
                self.annotation_path = None
                log.warning("No annotation path could be deduced from Saver object.")

    def output_annotations(
        self,
        annotation_path: Union[Path, str] = None,
    ) -> Path:
        """Output annotations to file.

        Args:
            annotation_path (Union[Path, str], optional): Output path for annotation file.

        Returns:
            Path: Path to annotation file.
        """
        if annotation_path is None:
            annotation_path = self.annotation_path
        log.info(f"Outputting annotation file to {annotation_path}")
        annotation_path = zpy.files.verify_path(annotation_path)
        return annotation_path



"""
    COCO (Common Objects in Context) dataset format.
"""
import copy
import logging
from pathlib import Path
from typing import List, Union

import gin
import zpy

log = logging.getLogger(__name__)


class COCOParseError(Exception):
    """Invalid COCO Annotation found when parsing data contents."""

    pass


@gin.configurable
class OutputCOCO(zpy.output.Output):
    """Output class for COCO style annotations.

    https://cocodataset.org/#home

    """

    ANNOTATION_FILENAME = Path("_annotations.coco.json")

    def __init__(self, *args, **kwargs) -> Path:
        super().__init__(*args, annotation_filename=self.ANNOTATION_FILENAME, **kwargs)

    @gin.configurable
    def output_annotations(
        self,
        annotation_path: Union[Path, str] = None,
        splitseg: bool = False,
    ) -> Path:
        """Output COCO annotations to file.

        Args:
            annotation_path (Union[Path, str], optional): Output path for annotation file.
            splitseg (bool, optional): Optionally output split-segmentation annotations. Defaults to False.

        Returns:
            Path: Path to annotation file.
        """
        annotation_path = super().output_annotations(annotation_path=annotation_path)
        coco_dict = {
            "info": self.coco_info(),
            "licenses": self.coco_license(),
            "categories": self.coco_categories(),
            "images": self.coco_images(),
            "annotations": self.coco_annotations(),
        }
        # Write out annotations to file
        zpy.files.write_json(annotation_path, coco_dict)
        parse_coco_annotations(annotation_path, data_dir=self.saver.output_dir)
        # Output split-segmentation annotations
        if splitseg:
            log.info(
                "Outputting COCO annotations with multi-part"
                + "segmentation split into seperate annotations"
            )
            coco_dict["annotations"] = self.coco_split_segmentation_annotations()
            annotation_path = zpy.files.add_to_path(annotation_path, "splitseg")
            # Write out annotations to file
            zpy.files.write_json(annotation_path, coco_dict)
            parse_coco_annotations(annotation_path, data_dir=self.saver.output_dir)
        return annotation_path

    @gin.configurable
    def coco_info(self, keys_to_add: List[str] = None):
        """coco info"""
        coco_info = {
            "description": self.saver.metadata["description"],
            "url": self.saver.metadata["url"],
            "version": self.saver.metadata["date_created"],
            "year": self.saver.metadata["year"],
            "contributor": self.saver.metadata["contributor"],
            "date_created": self.saver.metadata["date_created"],
            "save_path": self.saver.metadata["save_path"],
        }
        # Add any extra keys.
        if keys_to_add is not None:
            for key in keys_to_add:
                value = self.saver.metadata.get(key, None)
                if value is not None:
                    coco_info[key] = value
        return coco_info

    def coco_license(self):
        """coco license"""
        return {
            "url": "http://zumolabs.ai/image_license/",
            "id": 0,
            "name": "Zumo Labs Image License",
        }

    @gin.configurable
    def coco_categories(
        self,
        keys_to_add: List[str] = [
            "keypoints",
            "skeleton",
            "supercategory",
            "subcategories",
        ],
    ):
        """coco categories"""
        coco_categories = []
        for category in self.saver.categories.values():
            coco_category = {
                "id": category["id"],
                "name": category["name"],
            }
            # Add any extra keys.
            if keys_to_add is not None:
                for key in keys_to_add:
                    value = category.get(key, None)
                    if value is not None:
                        coco_category[key] = value
            coco_categories.append(coco_category)
        return coco_categories

    @gin.configurable
    def coco_images(
        self,
        only_default_images: bool = True,
        keys_to_add: List[str] = None,
    ):
        """coco images"""
        coco_images = []
        for image in self.saver.images.values():
            if only_default_images and not image["style"] == "default":
                # COCO annotations only have image annotations
                # for RGB images. No segmentation images.
                continue
            coco_img = {
                "license": 0,
                "id": image["id"],
                "file_name": image["name"],
                "coco_url": image["name"],
                "width": image["width"],
                "height": image["height"],
                "date_captured": self.saver.metadata["date_created"],
                "flickr_url": ".",
            }
            # Add any extra keys.
            if keys_to_add is not None:
                for key in keys_to_add:
                    value = image.get(key, None)
                    if value is not None:
                        coco_img[key] = value
            coco_images.append(coco_img)
        return coco_images

    @gin.configurable
    def coco_annotations(
        self,
        keys_to_add: List[str] = ["bbox", "area", "segmentation"],
        clipped: bool = True,
        only_default_images: bool = True,
    ):
        """coco annotations"""
        coco_annotations = []
        for annotation in self.saver.annotations:
            if (
                only_default_images
                and not self.saver.images[annotation["image_id"]]["style"] == "default"
            ):
                # COCO annotations only have image annotations
                # for RGB images. No segmentation images.
                continue
            coco_ann = {
                "category_id": annotation["category_id"],
                "image_id": annotation["image_id"],
                "id": annotation["id"],
                "iscrowd": False,
            }
            if clipped:
                height = self.saver.images[annotation["image_id"]]["height"]
                width = self.saver.images[annotation["image_id"]]["width"]

            # Add any extra keys.
            if keys_to_add is not None:
                for key in keys_to_add:
                    value = annotation.get(key, None)
                    if value is not None:
                        if key == "segmentation":
                            coco_ann["segmentation"] = (
                                self.saver.clip_coordinate_list(
                                    width=width,
                                    height=height,
                                    annotation=annotation["segmentation"],
                                )
                                if clipped
                                else annotation["segmentation"]
                            )
                        elif key == "segmentation_rle":
                            coco_ann["segmentation_rle"] = annotation[
                                "segmentation_rle"
                            ]
                        elif key == "segmentation_float":
                            coco_ann["segmentation_float"] = (
                                self.saver.clip_coordinate_list(
                                    normalized=True,
                                    annotation=annotation["segmentation_float"],
                                )
                                if clipped
                                else annotation["segmentation_float"]
                            )
                        elif key == "bbox_float":
                            coco_ann["bbox_float"] = (
                                self.saver.clip_bbox(
                                    normalized=True, bbox=annotation["bbox_float"]
                                )
                                if clipped
                                else annotation["bbox_float"]
                            )
                        elif key == "bbox":
                            coco_ann["bbox"] = (
                                self.saver.clip_bbox(
                                    width=width, height=height, bbox=annotation["bbox"]
                                )
                                if clipped
                                else annotation["bbox"]
                            )
                        elif key == "bboxes_float":
                            coco_ann["bboxes_float"] = (
                                [
                                    self.saver.clip_bbox(normalized=True, bbox=bbox)
                                    for bbox in annotation["bboxes_float"]
                                ]
                                if clipped
                                else annotation["bboxes_float"]
                            )
                        elif key == "bboxes":
                            coco_ann["bboxes"] = (
                                [
                                    self.saver.clip_bbox(
                                        height=height, width=width, bbox=bbox
                                    )
                                    for bbox in annotation["bboxes"]
                                ]
                                if clipped
                                else annotation["bboxes"]
                            )
                        else:
                            coco_ann[key] = value
                    try:
                        if key == "area":
                            coco_ann["area"] = (
                                annotation["bbox"][2] * annotation["bbox"][3]
                            )
                        elif key == "areas":
                            coco_ann["areas"] = [
                                bbox[2] * bbox[3] for bbox in annotation["bboxes"]
                            ]
                    except Exception:
                        pass
            # HACK: Require bbox for an annotation
            if coco_ann.get("bbox", None) is not None:
                coco_annotations.append(coco_ann)

        return coco_annotations

    @gin.configurable
    def coco_split_segmentation_annotations(
        self,
        keys_to_add: List[str] = ["bbox", "area", "segmentation"],
        clipped: bool = True,
        only_default_images: bool = True,
    ):
        """coco annotations one per segmentation"""
        coco_annotations = []
        # Annotation id will be re-mapped
        annotation_id = 0
        for annotation in self.saver.annotations:
            if (
                only_default_images
                and not self.saver.images[annotation["image_id"]]["style"] == "default"
            ):
                # COCO annotations only have image annotations
                # for RGB images. No segmentation images.
                continue
            coco_ann = {
                "category_id": annotation["category_id"],
                "image_id": annotation["image_id"],
                "iscrowd": False,
            }
            if clipped:
                height = self.saver.images[annotation["image_id"]]["height"]
                width = self.saver.images[annotation["image_id"]]["width"]

            # Add any extra keys.
            if keys_to_add is not None:
                for key in keys_to_add:
                    value = annotation.get(key, None)
                    if value is not None:
                        coco_ann[key] = value

            # Annotations can be composed of multiple annotation components
            if annotation.get("segmentation") is not None:
                num_components = len(annotation["segmentation"])
            else:
                log.warning(
                    "Skipping annotation: split segmentation requires segmentaiton field."
                )
                continue

            # TODO: This can prolly be cleaned up?
            for i in range(num_components):
                _coco_ann = copy.deepcopy(coco_ann)
                try:
                    _coco_ann["segmentation"] = (
                        [
                            self.saver.clip_coordinate_list(
                                height=height,
                                width=width,
                                annotation=annotation["segmentation"][i],
                            )
                        ]
                        if clipped
                        else [annotation["segmentation"][i]]
                    )
                except Exception:
                    pass
                try:
                    _coco_ann["segmentation_rle"] = [annotation["segmentation_rle"][i]]
                except Exception:
                    pass
                try:
                    _coco_ann["segmentation_float"] = (
                        [
                            self.saver.clip_coordinate_list(
                                normalized=True,
                                annotation=annotation["segmentation_float"][i],
                            )
                        ]
                        if clipped
                        else [annotation["segmentation_float"][i]]
                    )
                except Exception:
                    pass
                try:
                    _coco_ann["bbox_float"] = (
                        self.saver.clip_bbox(
                            normalized=True, bbox=annotation["bboxes_float"][i]
                        )
                        if clipped
                        else annotation["bboxes_float"][i]
                    )
                except Exception:
                    pass
                try:
                    _coco_ann["bbox"] = (
                        self.saver.clip_bbox(
                            width=width, height=height, bbox=annotation["bboxes"][i]
                        )
                        if clipped
                        else annotation["bboxes"][i]
                    )
                except Exception:
                    pass
                try:
                    _coco_ann["area"] = annotation["areas"][i]
                except Exception:
                    pass

                # HACK: Require bbox for an annotation
                if _coco_ann.get("bbox", None) is not None:
                    _coco_ann["id"] = annotation_id
                    annotation_id += 1
                    coco_annotations.append(_coco_ann)

        return coco_annotations


@gin.configurable
def parse_coco_annotations(
    annotation_file: Union[Path, str],
    data_dir: Union[Path, str] = None,
    output_saver: bool = False,
    # Specify which keys to add to ImageSaver
    image_keys_to_add: List[str] = None,
) -> zpy.saver_image.ImageSaver:
    """Parse COCO annotations, optionally output a ImageSaver object.

    Args:
        annotation_file (Union[Path, str]): Path to annotation file.
        data_dir (Union[Path, str], optional): Directory containing data (images, video, etc).
        output_saver (bool, optional): Whether to return a Saver object or not. Defaults to False.
        image_keys_to_add (List[str], optional): Image dictionary keys to include when parsing COCO dict.

    Raises:
        COCOParseError: Various checks on annotations, categories, images

    Returns:
        zpy.saver_image.ImageSaver: Saver object for Image datasets.
    """
    log.info(f"Parsing COCO annotations at {annotation_file}...")
    # Check annotation file path
    annotation_file = zpy.files.verify_path(annotation_file)
    if data_dir is not None:
        data_dir = zpy.files.verify_path(data_dir, check_dir=True)
    else:
        # If no data_dir, assume annotation file is in the root folder.
        data_dir = annotation_file.parent

    # Check that categories, images, and annotations are not blank
    coco_annotations = zpy.files.read_json(annotation_file)
    images = coco_annotations["images"]
    if len(images) == 0:
        raise COCOParseError(f"no images found in {annotation_file}")
    categories = coco_annotations["categories"]
    if len(categories) == 0:
        raise COCOParseError(f"no categories found in {annotation_file}")
    annotations = coco_annotations["annotations"]
    if len(annotations) == 0:
        raise COCOParseError(f"no annotations found in {annotation_file}")
    log.info(
        f"images:{len(images)} categories:{len(categories)} annotations:{len(annotations)}"
    )

    # Optionally output a saver object.
    if output_saver:
        saver = zpy.saver_image.ImageSaver(
            output_dir=data_dir,
            annotation_path=annotation_file,
            description=coco_annotations["info"]["description"],
            clean_dir=False,
        )

    # Check Images
    log.info("Parsing images...")
    img_ids = []
    for img in images:
        # Image ID
        image_id = img["id"]
        if not isinstance(image_id, int):
            raise COCOParseError(f"image id {image_id} must be int.")
        if image_id in img_ids:
            raise COCOParseError(f"image id {image_id} already used.")
        img_ids.append(image_id)
        if image_id < 0:
            raise COCOParseError(f"invalid image id {image_id}")
        # Height and Width
        height, width = img["height"], img["width"]
        if not isinstance(height, int):
            raise COCOParseError(f"height {height} must be int.")
        if not isinstance(width, int):
            raise COCOParseError(f"width {width} must be int.")
        if height <= 0 or width <= 0:
            raise COCOParseError(f"width and height h:{height} w:{width} must be > 0")
        # Image Name
        filename = img["file_name"]
        if not isinstance(filename, str):
            raise COCOParseError(f"filename {filename} must be str.")
        image_path = data_dir / filename
        if not image_path.exists():
            raise COCOParseError(f"image path {image_path} does not exist")
        # COCO Path
        coco_url = img.get("coco_url", None)
        if coco_url is None:
            coco_url = filename
        coco_url = Path(coco_url)
        coco_path = data_dir / coco_url
        if not coco_path.exists():
            raise COCOParseError(f"coco url {coco_path} does not exist")
        # Save each image to ImageSaver object
        if output_saver:
            saver.images[image_id] = {
                "id": image_id,
                "name": filename,
                "output_path": str(coco_url),
                "height": height,
                "width": width,
                "style": "default",
            }
            # Add any extra keys.
            if image_keys_to_add is not None:
                for key in image_keys_to_add:
                    value = img.get(key, None)
                    if value is not None:
                        saver.images[image_id][key] = value

    # Check Categories
    log.info("Parsing categories...")
    cat_ids = []
    cat_names = []
    for category in categories:
        name, category_id = category["name"], category["id"]
        log.info(f"name:{name} id:{category_id}")
        # Category Name
        category_name = category["name"]
        if not isinstance(category_name, str):
            raise COCOParseError(f"category_name {category_name} must be str.")
        if category_name in cat_names:
            raise COCOParseError(f"category_name {category_name} already used")
        cat_names.append(category_name)
        # Category ID
        category_id = category["id"]
        if not isinstance(image_id, int):
            raise COCOParseError(f"category_id {category_id} must be int.")
        if category_id in cat_ids:
            raise COCOParseError(f"category id {category_id} already used")
        cat_ids.append(category_id)
        # Supercategories
        if category.get("supercategory", None) is not None:
            pass
        if category.get("supercategories", None) is not None:
            pass
        # Subcategories
        if category.get("subcategory", None) is not None:
            pass
        if category.get("subcategories", None) is not None:
            pass
        # Keypoints
        if category.get("keypoints", None) is not None:
            keypoints = category["keypoints"]
            log.info(f"{len(keypoints)} keypoints:{keypoints}")
            if category.get("skeleton", None) is None:
                raise COCOParseError(f"skeleton must be present with {keypoints}")
        # Save each category to ImageSaver object
        if output_saver:
            _category = saver.categories.get(category_id, None)
            if _category is None:
                saver.categories[category_id] = {}
            for key, value in category.items():
                saver.categories[category_id][key] = value

    # Check Annotations
    log.info("Parsing annotations...")
    ann_ids = []
    for annotation in annotations:
        # IDs
        image_id, category_id, annotation_id = (
            annotation["image_id"],
            annotation["category_id"],
            annotation["id"],
        )
        if image_id not in img_ids:
            raise COCOParseError(f"annotation img:{image_id} not in {img_ids}")
        if category_id not in cat_ids:
            raise COCOParseError(f"annotation cat:{category_id} not in {cat_ids}")
        if annotation_id in ann_ids:
            raise COCOParseError(f"annotation id:{annotation_id} already used")
        ann_ids.append(annotation_id)

        # Bounding Boxes
        bbox = annotation.get("bbox", None)
        if bbox is not None:
            pass

        # Keypoints
        keypoints = annotation.get("num_keypoints", None)
        if keypoints is not None:
            if "keypoints_xyv" in annotation:
                if (
                    len(annotation["keypoints_xyv"])
                    != int(annotation["num_keypoints"]) * 3
                ):
                    raise COCOParseError(
                        "keypoints_xyv not correct size {len(keypoints)}"
                    )
            if "keypoints_xyz" in annotation:
                if (
                    len(annotation["keypoints_xyz"])
                    != int(annotation["num_keypoints"]) * 3
                ):
                    raise COCOParseError(
                        "keypoints_xyz not correct size {len(keypoints)}"
                    )

        # Save each annotation to ImageSaver object
        if output_saver:
            saver.annotations.append(annotation)

    if output_saver:
        return saver
    else:
        return None



"""
    Asset management.
"""
import logging
import os
from pathlib import Path

import bpy

import zpy

log = logging.getLogger(__name__)


def get_asset_lib_path() -> Path:
    """Returns path to asset library location.

    Defaults to the directory of the blenderfile.

    Returns:
        Path: pathlib.Path object to library location.
    """
    assets_env_path = os.environ.get("ASSETS", None)
    if assets_env_path is None:
        log.warning("Could not find environment variable $ASSETS")
        return None
    else:
        assets_env_path = zpy.files.verify_path(assets_env_path, check_dir=True)
        log.info(f"Found assets path at {assets_env_path}")
        return assets_env_path


def script_template_dir() -> Path:
    """Path to the script templates for zpy addon.

    Returns:
        pathlib.Path: Path to script templates for zpy addon.
    """
    script_path = Path(bpy.utils.script_path_user())
    template_dir = script_path / "addons" / "zpy_addon" / "templates"
    return zpy.files.verify_path(template_dir, check_dir=True)


def hdri_dir() -> Path:
    """Path to the HDRI directory.

    Returns:
        pathlib.Path: Path to HDRI directory.
    """
    asset_path = zpy.assets.get_asset_lib_path()
    if asset_path is None:
        # Path to directory containing default Blender HDRIs (exr)
        _path = Path(bpy.utils.resource_path("LOCAL"))
        hdri_dir = _path / "datafiles" / "studiolights" / "world"
    else:
        hdri_dir = asset_path / "lib" / "hdris" / "1k"
    hdri_dir = zpy.files.verify_path(hdri_dir, check_dir=True)
    log.debug(f"Using HDRI directory at {hdri_dir}")
    return hdri_dir


def texture_dir() -> Path:
    """Path to the textures directory.

    Returns:
        pathlib.Path: Path to textures directory.
    """
    asset_path = zpy.assets.get_asset_lib_path()
    if asset_path is None:
        _path = Path(bpy.utils.script_path_user())
        texture_dir = _path / "addons" / "zpy_addon" / "assets"
    else:
        texture_dir = asset_path / "lib" / "textures" / "random_512p"
    texture_dir = zpy.files.verify_path(texture_dir, check_dir=True)
    log.debug(f"Using texture directory at {texture_dir}")
    return texture_dir



"""
    KDTree utilities for Blender Python.
"""
import logging
from typing import List, Tuple

import bpy
import gin
import mathutils
import numpy as np

import zpy

log = logging.getLogger(__name__)


def kdtree_from_collection(
    collections: List[bpy.types.Collection],
) -> mathutils.kdtree.KDTree:
    """Creates a KDTree of vertices from a collection of objects."""
    # First get the size of the objects (number of vertices)
    size = 0
    for obj in zpy.objects.for_obj_in_collections(collections):
        size += len(obj.data.vertices)
    # Then add them to a tree object
    kd = mathutils.kdtree.KDTree(size)
    insert_idx = 0
    for obj in zpy.objects.for_obj_in_collections(collections):
        for v in obj.data.vertices:
            world_coordinate_v = obj.matrix_world @ v.co
            kd.insert(world_coordinate_v, insert_idx)
            insert_idx += 1
    # Balancing is the most expensive operation
    kd.balance()
    return kd


@gin.configurable
def floor_occupancy(
    kdtree: mathutils.kdtree.KDTree,
    x_bounds: Tuple[float],
    y_bounds: Tuple[float],
    z_height: float = 0.0,
    num_voxels: int = 100,
) -> float:
    """Get occupancy percentage for floor (XY plane)."""
    log.info("Calculating floor occupancy ....")
    x_side_length = abs(x_bounds[1] - x_bounds[0])
    y_side_length = abs(y_bounds[1] - y_bounds[0])
    # Number of voxels determines number of points in each dimmension
    voxel_cube_side_length = ((x_side_length * y_side_length) / num_voxels) ** (1 / 2)
    num_points_x = x_side_length / voxel_cube_side_length
    num_points_y = y_side_length / voxel_cube_side_length
    # TODO: This can definitely be vectorized better
    x_space, x_step = np.linspace(*x_bounds, num=int(num_points_x), retstep=True)
    y_space, y_step = np.linspace(*y_bounds, num=int(num_points_y), retstep=True)
    occupancy_grid = np.zeros((int(num_points_x), int(num_points_y)))
    for x_idx, x in enumerate(x_space):
        for y_idx, y in enumerate(y_space):
            x = float(x)
            y = float(y)
            closest_point = kdtree.find((x, y, z_height))[0]
            if (closest_point.x > (x - x_step)) and (closest_point.x < (x + x_step)):
                if (closest_point.y > (y - y_step)) and (
                    closest_point.y < (y + y_step)
                ):
                    occupancy_grid[x_idx][y_idx] = 1.0
    log.info("... Done.")
    log.debug(f"Floor occupancy grid: {str(occupancy_grid)}")
    return float(np.mean(occupancy_grid.copy()))


@gin.configurable
def volume_occupancy(
    kdtree: mathutils.kdtree.KDTree,
    x_bounds: Tuple[float],
    y_bounds: Tuple[float],
    z_bounds: Tuple[float],
    num_voxels: int = 100,
) -> float:
    """Get occupancy percentage for volume."""
    log.info("Calculating volume occupancy ....")
    x_side_length = abs(x_bounds[1] - x_bounds[0])
    y_side_length = abs(y_bounds[1] - y_bounds[0])
    z_side_length = abs(z_bounds[1] - z_bounds[0])
    # Number of voxels determines number of points in each dimmension
    voxel_cube_side_length = (
        (x_side_length * y_side_length * z_side_length) / num_voxels
    ) ** (1 / 3)
    num_points_x = x_side_length / voxel_cube_side_length
    num_points_y = y_side_length / voxel_cube_side_length
    num_points_z = z_side_length / voxel_cube_side_length
    # TODO: This can definitely be vectorized better
    x_space, x_step = np.linspace(*x_bounds, num=int(num_points_x), retstep=True)
    y_space, y_step = np.linspace(*y_bounds, num=int(num_points_y), retstep=True)
    z_space, z_step = np.linspace(*z_bounds, num=int(num_points_z), retstep=True)
    occupancy_grid = np.zeros((int(num_points_x), int(num_points_y), int(num_points_z)))
    for x_idx, x in enumerate(x_space):
        for y_idx, y in enumerate(y_space):
            for z_idx, z in enumerate(z_space):
                x = float(x)
                y = float(y)
                z = float(z)
                closest_point = kdtree.find((x, y, z))[0]
                if (closest_point.x > (x - x_step)) and (
                    closest_point.x < (x + x_step)
                ):
                    if (closest_point.y > (y - y_step)) and (
                        closest_point.y < (y + y_step)
                    ):
                        if (closest_point.z > (z - z_step)) and (
                            closest_point.z < (z + z_step)
                        ):
                            occupancy_grid[x_idx][y_idx][z_idx] = 1.0
    log.info("... Done.")
    log.debug(f"Volume occupancy grid: {str(occupancy_grid)}")
    return float(np.mean(occupancy_grid.copy()))



"""
    Logging utilities.
"""
from pathlib import Path
from typing import List, Union
import numpy as np
import logging
import shutil

log = logging.getLogger(__name__)


def set_log_levels(
    level: str = None,
    modules: List[str] = [
        "zpy",
        "zpy_addon",
        "bpy.zpy_addon" "neuralyzer",
        "bender",
    ],
    log_format: str = "%(asctime)s: %(levelname)s %(filename)s] %(message)s",
) -> None:
    """Set logger levels for all zpy modules.

    Args:
        level (str, optional): log level in [info, debug, warning]. Defaults to logging.Info.
        modules (List[str], optional): Modules to set logging for.
            Defaults to [ 'zpy', 'zpy_addon', 'bpy.zpy_addon' 'neuralyzer', ].
        log_format (str, optional): Log format string.
            Defaults to '%(asctime)s: %(levelname)s %(filename)s] %(message)s'
    """
    if level is None:
        log_level = logging.INFO
    elif level == "info":
        log_level = logging.INFO
    elif level == "debug":
        log_level = logging.DEBUG
    elif level == "warning":
        log_level = logging.WARNING
    else:
        log.warning(f"Invalid log level {level}")
        return
    logging.basicConfig(format=log_format)
    for logger_name in modules:
        try:
            log.warning(f"Setting log level for {logger_name} to {log_level} ({level})")
            logging.getLogger(logger_name).setLevel(log_level)
        except Exception:
            pass


def linebreaker_log(
    message: str,
    line_length: int = 80,
):
    """Good looking line-breaker log message.

    Args:
        message (str): Message to put in the log.
        line_length (int, optional): Length of line breakers ----. Defaults to 80.
    """
    # Clip the message
    message = message[:line_length]
    whitespace = " " * int((line_length - len(message)) / 2)
    # La piece de resistance
    log.info("-" * line_length)
    log.info(f"{whitespace}{message.upper()}{whitespace}")
    log.info("-" * line_length)


def setup_file_handlers(
    log_dir: Union[str, Path] = "/tmp",
    error_log: bool = True,
    debug_log: bool = True,
) -> None:
    """Output log files for requests

    Args:
        error_log: output error.log
        debug_log: output debug.log
        log_dir: directory to output log files
    """
    root = logging.getLogger()

    info_fh = logging.FileHandler(f"{log_dir}/info.log", mode="w")
    info_fh.setLevel(logging.INFO)
    root.addHandler(info_fh)

    if error_log:
        error_fh = logging.FileHandler(f"{log_dir}/error.log", mode="w")
        error_fh.setLevel(logging.ERROR)
        root.addHandler(error_fh)

    if debug_log:
        debug_fh = logging.FileHandler(f"{log_dir}/debug.log", mode="w")
        debug_fh.setLevel(logging.DEBUG)
        root.addHandler(debug_fh)


def save_log_files(
    output_dir: Union[str, Path],
    log_dir: Union[str, Path] = "/tmp",
) -> None:
    """Save log files to output directory

    Args:
        output_dir: directory to save log files
        log_dir: directory where logs exist
    """
    for log in ["info.log", "debug.log", "error.log"]:
        src = Path(log_dir) / log
        dst = Path(output_dir) / log
        if src.exists() and src != dst:
            shutil.copy(src, dst)


def parse_log_file(log_file: Union[str, Path]) -> None:
    import re

    step_times, render_times = [], []
    with open(log_file, "r") as f:
        render_in_step = []
        for line in f.readlines():
            seconds = re.search("\d+\.\d+(?=s)", line)
            if line.startswith("Rendering took"):
                render_in_step.append(float(seconds.group(0)))
            elif line.startswith("Simulation step took"):
                render_times.append(render_in_step)
                render_in_step = []
                step_times.append(float(seconds.group(0)))

    return {
        "avg_step_time": sum(step_times) / len(step_times),
        "avg_render_time": [
            v / len(render_times) for v in np.sum(render_times, axis=0)
        ],
        "step_times": step_times,
        "render_times": render_times,
    }



"""
    HDRI utilities. You can find lots of free HDRIs at https://hdrihaven.com/
"""
import logging
import math
import random
from pathlib import Path
from typing import Tuple, Union

import bpy
import mathutils

import gin
import zpy

log = logging.getLogger(__name__)


@gin.configurable
def load_hdri(
    path: Union[Path, str],
    scale: Tuple[float] = (1.0, 1.0, 1.0),
    random_z_rot: bool = True,
) -> None:
    """Load an HDRI from path.

    Args:
        path (Union[Path, str]): Path to the HDRI.
        scale (Tuple[float], optional): Scale in (x, y, z). Defaults to (1.0, 1.0, 1.0).
        random_z_rot (bool, optional): Randomly rotate HDRI around Z axis. Defaults to True.
    """
    scene = zpy.blender.verify_blender_scene()
    scene.world.use_nodes = True
    tree = scene.world.node_tree
    out_node = zpy.nodes.get_or_make(
        "World Output", "ShaderNodeOutputWorld", tree, pos=(0, 0)
    )
    bg_node = zpy.nodes.get_or_make(
        "Background", "ShaderNodeBackground", tree, pos=(-150, 0)
    )
    env_node = zpy.nodes.get_or_make(
        "Environment Texture", "ShaderNodeTexEnvironment", tree, pos=(-400, 0)
    )
    log.info(f"Loading HDRI at {path}")
    path = zpy.files.verify_path(path, make=False)
    env_node.image = bpy.data.images.load(str(path))
    env_node.texture_mapping.scale = mathutils.Vector(scale)
    world_rot_node = zpy.nodes.get_or_make(
        "World Rotation", "ShaderNodeVectorRotate", tree, pos=(-550, 0)
    )
    world_rot_node.rotation_type = "Z_AXIS"
    if random_z_rot:
        world_rotation = random.uniform(0, math.pi)
        log.debug(f"Rotating HDRI randomly along Z axis to {world_rotation}")
        world_rot_node.inputs["Angle"].default_value = world_rotation
    texcoord_node = zpy.nodes.get_or_make(
        "Texture Coordinate", "ShaderNodeTexCoord", tree, pos=(-730, 0)
    )
    # Link all the nodes together
    tree.links.new(out_node.inputs[0], bg_node.outputs[0])
    tree.links.new(bg_node.inputs[0], env_node.outputs[0])
    tree.links.new(env_node.inputs[0], world_rot_node.outputs[0])
    tree.links.new(world_rot_node.inputs[0], texcoord_node.outputs[0])


@gin.configurable
def random_hdri(
    apply_to_scene: bool = True,
    scale: Tuple[float] = (1.0, 1.0, 1.0),
) -> Path:
    """Generate a random HDRI from an asset path.

    Args:
        apply_to_scene (bool, optional): Load the HDRI into the active scene. Defaults to True.

    Returns:
        Path: Path to the random HDRI.
    """
    hdri_dir = zpy.assets.hdri_dir()
    hdri_path = zpy.files.pick_random_from_dir(
        hdri_dir, suffixes=[".exr", ".hdri", ".hdr"]
    )
    if apply_to_scene:
        load_hdri(hdri_path, scale=scale)
    return hdri_path



from cli.utils import fetch_auth
import requests
import json


@fetch_auth
def create_transform(name, operation, config, project, url, auth_headers):
    """create transform

    Create a job object on ZumoLabs backend. This will trigger the backend
    to run the job.

    Args:
        name (str): name of dataset
        operation (str): transform type
        config (dict): configuration for transform
        project (str): project uuid
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    from cli.datasets import fetch_dataset

    dataset = fetch_dataset(name)
    endpoint = f"{url}/api/v1/transforms/"
    data = {
        "project": project,
        "operation": operation,
        "name": name,
        "input_dataset": dataset["id"],
        "config": json.dumps(config),
    }
    r = requests.post(endpoint, data=data, headers=auth_headers)
    if r.status_code != 201:
        r.raise_for_status()


@fetch_auth
def fetch_transforms(filters, url, auth_headers):
    """fetch transforms

    Fetch transform objects from ZumoLabs backend.

    Args:
        filters (dict): query param filters for API call
        url (str): backend endpoint
        auth_headers: authentication for backend

    Returns:
        list: list of transforms
    """
    endpoint = f"{url}/api/v1/transforms/"
    r = requests.get(endpoint, headers=auth_headers, params=filters)
    if r.status_code != 200:
        r.raise_for_status()
    return json.loads(r.text)["results"]


@fetch_auth
def available_transforms(url, auth_headers):
    """available transforms

    List all transforms available on the backend.

    Args:
        url (str): backend endpoint
        auth_headers: authentication for backend

    Returns:
        list: list of transforms
    """
    endpoint = f"{url}/api/v1/transforms/available/"
    r = requests.get(endpoint, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    return json.loads(r.text)



"""
    Utilities for color.
"""
import logging
from pathlib import Path
from typing import List, Tuple, Union

import zpy

log = logging.getLogger(__name__)

# Random colors are loaded on module import
RANDOM_COLOR_IDX = 1
COLORS_FILE = "colors.json"
COLORS = None


def reset(random_color_idx: int = 1):
    """Load colors from file and reset random idx."""
    global COLORS, RANDOM_COLOR_IDX
    _path = Path(__file__).parent / COLORS_FILE
    zpy.files.verify_path(_path)
    COLORS = zpy.files.read_json(_path)
    RANDOM_COLOR_IDX = random_color_idx


def hex_to_irgb(hex_value: str) -> Tuple[int]:
    """Convert hex value to integer rgb (0 to 255)."""
    hex_value = int(hex_value[1:], 16)
    b = hex_value & 0xFF
    g = (hex_value >> 8) & 0xFF
    r = (hex_value >> 16) & 0xFF
    return r, g, b


def frgb_to_frgba(rgb: Tuple[float], a=1.0) -> Tuple[float]:
    """Convert 3-channel RGB to 4-channel RGBA."""
    _val = *rgb, a
    return _val


def hex_to_frgb(hex_value: str) -> Tuple[float]:
    """Convert hex value to float rgb (0 to 1)."""
    return irgb_to_frgb(hex_to_irgb(hex_value))


def irgb_to_frgb(irgb: Tuple[int]) -> Tuple[float]:
    """Convert integer rgb (0 to 255) to float rgb (0 to 1)."""
    max_rgb_value = 255
    return tuple((x / max_rgb_value) for x in irgb)


def irgb_to_hex(irgb: Tuple[int]) -> str:
    """Convert integer rgb (0 to 255) to hex."""
    r, g, b = irgb
    return "#%02x%02x%02x" % (r, g, b)


def frgb_to_irgb(frgb: Tuple[float]) -> Tuple[int]:
    """Convert float rgb (0 to 1) to integer rgb (0 to 255)."""
    max_rgb_value = 255
    return tuple((int(x * max_rgb_value) for x in frgb))


def frgb_to_hex(frgb: Tuple[float]) -> str:
    """Convert float rgb (0 to 1) to hex."""
    return irgb_to_hex(frgb_to_irgb(frgb))


def _gamma_correction(channel: float) -> float:
    """Convert float rgb (0 to 1) to the gamma-corrected sRGB float (0 to 1)
    for a single color channel."""
    exponent_srgb = 1 / 2.2
    return channel ** exponent_srgb


def frgb_to_srgb(frgb: Tuple[float]) -> Tuple[float]:
    """Convert float rgb (0 to 1) to the gamma-corrected sRGB float (0 to 1)."""
    return tuple(_gamma_correction(x) for x in frgb)


def frgb_to_srgba(frgb: Tuple[float], a=1.0) -> Tuple[float]:
    """Convert float rgb (0 to 1) to the gamma-corrected sRGBA float (0 to 1)."""
    srgb = frgb_to_srgb(frgb)
    srgba = frgb_to_frgba(srgb, a=a)
    return srgba


def _output_style(
    name: str, hex: str, output_style: str
) -> Union[Tuple[float, int, str], str]:
    """Convert hex to an output style."""
    if output_style == "frgb":
        return hex_to_frgb(hex)
    if output_style == "frgba":
        return frgb_to_frgba(hex_to_frgb(hex))
    elif output_style == "irgb":
        return hex_to_irgb(hex)
    elif output_style == "hex":
        return hex
    elif output_style == "name_irgb":
        return name, hex_to_irgb(hex)
    elif output_style == "name_frgb":
        return name, hex_to_frgb(hex)
    elif output_style == "name_frgba":
        return name, frgb_to_frgba(hex_to_frgb(hex))
    else:
        raise ValueError("Color must be frgb, irgb, or hex.")


def default_color(output_style: str = "frgb") -> Union[Tuple[float, int, str], str]:
    """Default color."""
    global COLORS
    if COLORS is None:
        reset()
    _name = COLORS[0]["name"]
    _hex = COLORS[0]["hex"]
    log.debug(f"Default color chosen is {_name} - {_hex} - {hex_to_frgb(_hex)}")
    return _output_style(_name, _hex, output_style=output_style)


def random_color(output_style: str = "frgb") -> Union[Tuple[float, int, str], str]:
    """Random color.

    This will go through a pre-baked list every time,
    to prevent different random seeds from changing the
    color for a category.

    """
    global RANDOM_COLOR_IDX, COLORS
    if COLORS is None:
        reset()
    _name = COLORS[RANDOM_COLOR_IDX]["name"]
    _hex = COLORS[RANDOM_COLOR_IDX]["hex"]
    # Update global color idx
    RANDOM_COLOR_IDX += 1
    if RANDOM_COLOR_IDX > len(COLORS):
        log.error("Ran out of unique colors!")
    log.debug(f"Random color chosen is {_name} - {_hex} - {hex_to_frgb(_hex)}")
    return _output_style(_name, _hex, output_style=output_style)


def closest_color(
    color: Tuple[float],
    colors: List[Tuple[float]],
    max_dist: float = 0.01,
) -> Union[None, Tuple[float]]:
    """Get the index of the closest color in a list to the input color."""
    min_dist = 3.0
    nearest_idx = 0
    for i, _color in enumerate(colors):
        dist = (
            (color[0] - _color[0]) ** 2
            + (color[1] - _color[1]) ** 2
            + (color[2] - _color[2]) ** 2
        )
        if dist < min_dist:
            min_dist = dist
            nearest_idx = i
    if min_dist > max_dist:
        log.debug(f"No color close enough w/ maxmimum distance of {max_dist}")
        return None
    return colors[nearest_idx]



from cli.utils import fetch_auth, download_url
from zpy.files import to_pathlib_path
import json
import requests

LOG_TYPES = ["info", "debug", "error"]


@fetch_auth
def fetch_logs(resource, name, path, url, auth_headers):
    """fetch logs

    Fetch LOG_TYPES for a backend run.

    Args:
        type (str): resource to fetch logs
        name (str): name of resource
        path (str): output_dir
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    endpoint = f"{url}/api/v1/{resource}/"
    r = requests.get(endpoint, params={"name": name}, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    response = json.loads(r.text)
    if response["count"] != 1:
        raise NameError(f"found {response['count']} {resource} for name {name}")
    obj = response["results"][0]
    endpoint = f"{url}/api/v1/{resource}/{obj['id']}/logs"
    r = requests.get(endpoint, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    response = json.loads(r.text)
    output_dir = to_pathlib_path(path)
    for log_type in LOG_TYPES:
        output_path = output_dir / f"{log_type}.log"
        download_url(response[log_type]["redirect_link"], output_path)



import json

import requests

from cli.utils import fetch_auth


@fetch_auth
def fetch_accounts(filters, url, auth_headers):
    """fetch accounts

    Fetch accounts from ZumoLabs backend.

    Args:
        filters (dict): query param filters for API call
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    endpoint = f"{url}/api/v1/accounts/"
    r = requests.get(endpoint, headers=auth_headers, params=filters)
    if r.status_code != 200:
        r.raise_for_status()
    return json.loads(r.text)["results"]



import functools
import math
from copy import deepcopy
from itertools import product
from urllib.request import urlopen

import click
from tqdm import tqdm

from cli.config import read_config


def parse_filter(str_filter):
    """parse filter

    Parse filter string to field, pattern, regex.
    Valid patterns depend on field but can include exact, startswith, contains,
    iexact, istartswith, icontains. If no pattern or field is provided the defaults
    are name for field and startswith for pattern.

    Examples:
        icontains:hello
        name:icontains:foobar
        state:iexact:ready
        foo

    Args:
        filter (str): filter query

    Return:
        field: field to filter on
        pattern: pattern to apply regex
        regex: string regex for pattern
    """
    filter_arr = str_filter.split(":")
    field, pattern, regex = "name", "startswith", filter_arr[-1]
    if len(filter_arr) == 2:
        pattern = filter_arr[0]
    elif len(filter_arr) == 3:
        field, pattern = filter_arr[0], filter_arr[1]
    return field, pattern, regex


def resolve_sweep(sweep_config):
    """resolve sweep

    Resolve a dictionary into a sweep of dictionaries. Gin bindings
    are used to communicate with job code and therefore the sweep is done
    across gin_bindings in the sweep_config.

    Example:
        {'foo': ['a', 'b']} -> {'foo': 'a'} {'foo': 'b'}

    Args:
        sweep_config (dict): dictionary to unfold into sweep

    Returns:
        list: list of configs resolved from sweep config
    """
    configs, bindings = [], sweep_config["gin_bindings"]
    for random_binding in [dict(zip(bindings, v)) for v in product(*bindings.values())]:
        config = deepcopy(sweep_config)
        config["gin_bindings"] = random_binding
        configs.append(config)
    return configs


def parse_args(args):
    """parse args

    Used by cli to parse arguments passed to cli calls. Includes
    safe eval function to convert from string to other types.

    Example:
        foo 1 bar 2 -> {'foo': 1, 'bar': 2}
    """

    def _safe_eval(key):
        try:
            return eval(key)
        except Exception:
            return key

    keys = args[::2]
    vals = map(lambda x: _safe_eval(x), args[1::2])
    return dict(zip(keys, vals))


def download_url(url, output_path):
    """download url

    Download from url to give output path and visualize using tqdm.

    Args:
        url (str): url to download
        output_path (str): path to download file to
    """
    u = urlopen(url)
    h = u.info()
    totalSize = int(h["Content-Length"])

    fp = open(output_path, "wb")

    blockSize = 8192
    with tqdm(total=totalSize) as pbar:
        while True:
            chunk = u.read(blockSize)
            if not chunk:
                break
            fp.write(chunk)
            pbar.update(blockSize)

    fp.flush()
    fp.close()


def fetch_auth(func):
    """fetch authentication

    Decorator to wrap functions providing the backend url and the
    correct authorization headers for requests.

    Args:
        func: function to wrap

    Returns:
        wrapped function
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        config = read_config()
        endpoint = config["ENDPOINT"]

        auth_header = {"Authorization": "token {}".format(config["TOKEN"])}
        return func(*args, **kwargs, url=endpoint, auth_headers=auth_header)

    return wrapper


def use_project(required=False):
    def use_project_inner(func):
        """Inject project uuid into function call. Optionally throw an error if it has not been set.

        Args:
            func: function to wrap

        Returns:
            wrapped function
        """

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            config = read_config()
            project = config.get("PROJECT", None)
            if project:
                click.echo(f"Using project {project}")
                return func(*args, **kwargs, project=project)
            else:
                if required:
                    click.secho(
                        "Project is not set. See `zpy project --help`",
                        fg="red",
                        err=True,
                    )
                    return

                return func(*args, **kwargs)

        return wrapper

    return use_project_inner


def print_list_as_columns(list_of_strings, num_cols=5, indent_prefix="    "):
    """Format and echo a list of strings into nicely formatted columns.

    Args:
        list_of_strings (list of str): A list of similar strings to format into columns.
        num_cols (int): Desired number of columns.
        indent_prefix (str): String to attach to the beginning of every printed line.
    Returns:
        None
    """
    count = len(list_of_strings)
    col_width = max(len(string) for string in list_of_strings)
    num_rows = math.ceil(count / num_cols)
    for i in range(num_rows):
        start_index = i * num_cols
        end_index = (i + 1) * num_cols
        if end_index > len(list_of_strings):
            end_index = len(list_of_strings)
        row = list_of_strings[start_index:end_index]

        format_string = indent_prefix + " ".join(
            ["{{:<{}}}".format(col_width) for _ in row]
        )

        click.echo(format_string.format(*row))



"""
    CSV (comma separated value) dataset format.
"""
import logging
from pathlib import Path
from typing import Callable, List, Union

import gin
import zpy

log = logging.getLogger(__name__)


class CSVParseError(Exception):
    """Invalid CSV Annotation found when parsing data contents."""

    pass


@gin.configurable
class OutputCSV(zpy.output.Output):
    """Output class for CSV (comma separated value) style annotations."""

    ANNOTATION_FILENAME = Path("annotations.csv")

    def __init__(self, *args, **kwargs) -> Path:
        super().__init__(*args, annotation_filename=self.ANNOTATION_FILENAME, **kwargs)

    @gin.configurable
    def output_annotations(
        self,
        annotation_path: Union[Path, str] = None,
        annotation_dict_to_csv_row_func: Callable = None,
        header: List[str] = None,
    ) -> Path:
        """Output CSV annotations to file.

        Args:
            annotation_path (Union[Path, str], optional): Output path for annotation file.
            annotation_dict_to_csv_row_func (Callable, optional): Function that converts an annotation dict to csv row.
                Defaults to None.
            header (List[str], optional): Column headers. Defaults to None.

        Returns:
            Path: Path to annotation file.
        """
        annotation_path = super().output_annotations(annotation_path=annotation_path)
        if annotation_dict_to_csv_row_func is None:
            raise CSVParseError(
                "Output CSV annotations requires a annotation_dict_to_csv_row_func"
            )
        csv_data = []
        if header is not None:
            csv_data.append(header)
        for annotation in self.saver.annotations:
            row = annotation_dict_to_csv_row_func(annotation, saver=self.saver)
            if row is not None:
                csv_data.append(row)
        # Write out annotations to file
        zpy.files.write_csv(annotation_path, csv_data)
        # Verify annotations
        parse_csv_annotations(annotation_path)
        return annotation_path


@gin.configurable
def parse_csv_annotations(
    annotation_file: Union[Path, str],
) -> None:
    """Parse CSV annotations.

    Args:
        annotation_file (Union[Path, str]): Path to annotation file.

    Raises:
        CSVParseError: Rows not same length.
    """
    log.info(f"Verifying CSV annotations at {annotation_file}...")
    csv_data = zpy.files.read_csv(annotation_file)
    # Make sure all the rows are the same length
    csv_data_iterable = iter(csv_data)
    try:
        length = len(next(csv_data_iterable))
    except StopIteration:
        raise CSVParseError(f"No data found in CSV at {annotation_file}")
    log.debug(f"Row length in CSV: {[length for l in csv_data_iterable]}")
    if not all(len(row) == length for row in csv_data_iterable):
        raise CSVParseError(f"Not all rows in the CSV have same length {length}")
    # TODO: Return Saver object.



"""
    Object utilities for Blender Python.
"""
import logging
import random
from pathlib import Path
from typing import List, Tuple, Union

import bpy
import gin
import mathutils
import zpy

log = logging.getLogger(__name__)


def verify(
    obj: Union[bpy.types.Object, str],
    check_none=True,
) -> bpy.types.Object:
    """Return object given name or Object type object.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
        check_none (bool, optional): Raise error if object does not exist. Defaults to True.

    Raises:
        ValueError: Object does not exist.

    Returns:
        bpy.types.Object: Scene object.
    """
    if isinstance(obj, str):
        obj = bpy.data.objects.get(obj)
    if check_none and obj is None:
        raise ValueError(f"Could not find object {obj}.")
    return obj


def load_blend_obj(
    name: str,
    path: Union[Path, str],
    link: bool = False,
) -> bpy.types.Object:
    """Load object from blend file.

    Args:
        name (str): Name of object to be loaded.
        path (Union[Path, str]): Path to the blender file with the object.
        link (bool, optional): Whether to link object to scene. Defaults to False.

    Returns:
        bpy.types.Object: Scene object that was loaded in.
    """
    path = zpy.files.verify_path(path, make=False)
    scene = zpy.blender.verify_blender_scene()
    with bpy.data.libraries.load(str(path), link=link) as (data_from, data_to):
        for from_obj in data_from.objects:
            if from_obj.startswith(name):
                log.debug(f"Loading obj {from_obj} from {str(path)}.")
                data_to.objects.append(from_obj)
    # Copy objects over to the current scene
    for obj in data_to.objects:
        scene.collection.objects.link(obj)
    for texture_folder_name in ["Textures", "textures", "TEX"]:
        texture_dir = path.parent / texture_folder_name
        if texture_dir.exists():
            bpy.ops.file.find_missing_files(directory=str(texture_dir))
            break
    return bpy.data.objects[name]


def select(
    obj: Union[bpy.types.Object, str],
) -> None:
    """Select an object.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
    """
    obj = verify(obj)
    log.debug(f"Selecting obj: {obj.name}")
    view_layer = zpy.blender.verify_view_layer()
    log.debug(f"Before select, bpy.context.active_object = {bpy.context.active_object}")
    log.debug(f"Before select, view_layer.objects.active = {view_layer.objects.active}")
    # De-select everything
    bpy.ops.object.select_all(action="DESELECT")
    if bpy.context.active_object is not None:
        bpy.context.active_object.select_set(False)
    view_layer.objects.active = None
    # Select the new object
    view_layer.objects.active = obj
    bpy.context.view_layer.objects.active = obj
    bpy.data.objects[obj.name].select_set(True, view_layer=view_layer)
    log.debug(f"After select, bpy.context.active_object = {bpy.context.active_object}")
    log.debug(f"After select, view_layer.objects.active = {view_layer.objects.active}")


def delete_obj(
    obj: Union[bpy.types.Object, str],
) -> None:
    """Delete an object.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
    """
    obj = verify(obj)
    select(obj)
    log.debug(f"Removing obj: {obj.name}")
    # bpy.ops.object.delete()
    bpy.data.objects.remove(obj, do_unlink=True)


def delete_obj_context(
    obj: Union[bpy.types.Object, str],
) -> None:
    """Alternative way to delete an object.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
    """
    obj = verify(obj)
    log.debug(f"Removing obj: {obj.name}")
    context_remove = bpy.context.copy()
    context_remove["selected_objects"] = [obj]
    bpy.ops.object.delete(context_remove)


def empty_collection(
    collection: bpy.types.Collection = None,
    method: str = "data",
) -> None:
    """Delete all objects in a collection

    Args:
        collection (bpy.types.Collection, optional): Optional collection to put new object inside of. Defaults to None.
        method (str, optional): Deletetion method, the values are data and context
    """
    if collection and (collection in list(bpy.data.collections)):
        if method == "data":
            for obj in collection.all_objects:
                bpy.data.objects.remove(obj, do_unlink=True)
        elif method == "context":
            context_remove = bpy.context.copy()
            context_remove["selected_objects"] = collection.all_objects
            bpy.ops.object.delete(context_remove)


def is_inside(
    location: Union[Tuple[float], mathutils.Vector],
    obj: Union[bpy.types.Object, str],
) -> bool:
    """Is point inside a mesh.

    https://blender.stackexchange.com/questions/31693/how-to-find-if-a-point-is-inside-a-mesh

    Args:
        location (Union[Tuple[float], mathutils.Vector]): Location (3-tuple or Vector) of point in 3D space.
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)

    Returns:
        bool: Whether object is inside mesh.
    """
    if not isinstance(location, mathutils.Vector):
        location = mathutils.Vector(location)
    is_found, closest_point, normal, _ = obj.closest_point_on_mesh(location)
    if not is_found:
        return False
    p2 = closest_point - location
    v = p2.dot(normal)
    return not (v < 0.0)


def for_obj_in_selected_objs(context) -> bpy.types.Object:
    """Safe iterable for selected objects.

    Yields:
        bpy.types.Object: Objects in selected objects.
    """
    zpy.blender.verify_view_layer()
    for obj in context.selected_objects:
        # Only meshes or empty objects TODO: Why the empty objects
        if not (obj.type == "MESH" or obj.type == "EMPTY"):
            continue
        # Make sure object exists in the scene
        if bpy.data.objects.get(obj.name, None) is None:
            continue
        context.view_layer.objects.active = obj
        yield obj


def for_obj_in_collections(
    collections: List[bpy.types.Collection],
    filter_mesh: bool = False,
) -> bpy.types.Object:
    """Yield objects in list of collection.

    Yields:
        bpy.types.Object: Object in collection.
    """
    for collection in collections:
        # TODO: Windows does not like this
        if not len(collection.objects) > 0:
            log.debug(f"Collection {collection.name} is empty, skipping...")
            continue
        for obj in collection.all_objects:
            if filter_mesh and obj.type == "MESH":
                # This gives you direct access to data block
                yield bpy.data.objects[obj.name]
            else:
                yield bpy.data.objects[obj.name]


def toggle_hidden(
    obj: Union[bpy.types.Object, str],
    hidden: bool = True,
    filter_string: str = None,
) -> None:
    """Recursive function to make object and children invisible.

    Optionally filter by a string in object name.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
        hidden (bool, optional): Whether to hide or un-hide object. Defaults to True.
        filter_string (str, optional): Filter objects to hide based on name containing this string. Defaults to None.
    """
    obj = verify(obj)
    if hasattr(obj, "hide_render") and hasattr(obj, "hide_viewport"):
        if (filter_string is None) or (filter_string in obj.name):
            log.debug(f"Hiding object {obj.name}")
            bpy.data.objects[obj.name].select_set(True)
            bpy.data.objects[obj.name].hide_render = hidden
            bpy.data.objects[obj.name].hide_viewport = hidden
        else:
            log.debug(
                f"Object {obj.name} does not contain filter string {filter_string}"
            )
    else:
        log.warning("Object does not have hide properties")
        return
    for child in obj.children:
        toggle_hidden(child, hidden=hidden, filter_string=filter_string)


def randomly_hide_within_collection(
    collections: List[bpy.types.Collection],
    chance_to_hide: float = 0.9,
) -> None:
    """Randomly hide objects in a list of collections.

    Args:
        collections (List[bpy.types.Collection]): A scene collection.
        chance_to_hide (float, optional): Probability of hiding an object in the collection. Defaults to 0.9.
    """
    to_hide = []
    for obj in for_obj_in_collections(collections):
        if random.random() < chance_to_hide:
            to_hide.append(obj.name)
    # HACK: hide objects by name, this causes segfault
    # if done in the for loop above, due to some kind of
    # pass by reference vs by value shenaniganry going on
    # with blender python sitting on top of blender C
    for name in to_hide:
        bpy.data.objects[name].select_set(True)
        bpy.data.objects[name].hide_render = True
        bpy.data.objects[name].hide_viewport = True


def segment(
    obj: Union[bpy.types.Object, str],
    name: str = "default",
    color: Tuple[float] = None,
    as_category: bool = False,
    as_single: bool = False,
) -> None:
    """Segment an object.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
        name (str, optional): Name of category or instance. Defaults to 'default'.
        color (Tuple[float], optional): Segmentation color. Defaults to None.
        as_category (bool, optional): Segment as a category, if false will segment as instance. Defaults to False.
        as_single (bool, optional): Segment all child objects as well. Defaults to False.
    """
    if "use_sculpt_vertex_colors" in dir(bpy.context.preferences.experimental):
        bpy.context.preferences.experimental.use_sculpt_vertex_colors = True
    obj = verify(obj)
    if color is None:
        color = zpy.color.random_color(output_style="frgb")
    obj.color = zpy.color.frgb_to_frgba(color)
    if as_category:
        obj.seg.category_name = name
        obj.seg.category_color = color
        seg_type = "category"
    else:
        obj.seg.instance_name = name
        obj.seg.instance_color = color
        seg_type = "instance"
    # Make sure object material is set up correctly with AOV nodes
    populate_vertex_colors(obj, zpy.color.frgb_to_frgba(color), seg_type)
    zpy.material.make_aov_material_output_node(obj=obj, style=seg_type)
    # Recursively add property to children objects
    if as_single:
        for child in obj.children:
            segment(
                obj=child,
                name=name,
                color=color,
                as_category=as_category,
                as_single=as_single,
            )


def populate_vertex_colors(
    obj: Union[bpy.types.Object, str],
    color_rgba: Tuple[float],
    seg_type: str = "instance",
) -> None:
    """Fill the given Vertex Color Layer with the color parameter values.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
        color_rgba (Tuple[float]): Segmentation color.
        seg_type (str, optional): Instance or Category segmentation. Defaults to 'instance'.
    """
    obj = verify(obj)
    if not obj.type == "MESH":
        log.warning(f"Object {obj.name} is not a mesh, has no vertices.")
        return
    # TODO: Is this select needed?
    # select(obj)
    # Remove any existing vertex color data
    if len(obj.data.sculpt_vertex_colors):
        for vcol in obj.data.sculpt_vertex_colors.keys():
            if seg_type in vcol:
                obj.data.sculpt_vertex_colors.remove(
                    obj.data.sculpt_vertex_colors[seg_type]
                )
    # Add new vertex color data
    obj.data.sculpt_vertex_colors.new(name=seg_type)
    # Iterate through each vertex in the mesh
    for i, _ in enumerate(obj.data.vertices):
        obj.data.sculpt_vertex_colors[seg_type].data[i].color = color_rgba


def random_position_within_constraints(
    obj: Union[bpy.types.Object, str],
) -> None:
    """Randomize position of object within constraints.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
    """
    obj = verify(obj)
    # Make sure object has constraints
    _constraints = obj.constraints.get("Limit Location", None)
    if _constraints is not None:
        obj.location.x = random.uniform(
            obj.constraints["Limit Location"].min_x,
            obj.constraints["Limit Location"].max_x,
        )
        obj.location.y = random.uniform(
            obj.constraints["Limit Location"].min_y,
            obj.constraints["Limit Location"].max_y,
        )
        obj.location.z = random.uniform(
            obj.constraints["Limit Location"].min_z,
            obj.constraints["Limit Location"].max_z,
        )


@gin.configurable
def copy(
    obj: Union[bpy.types.Object, str],
    name: str = None,
    collection: bpy.types.Collection = None,
    is_library_object: bool = False,
    is_copy: bool = True,
) -> bpy.types.Object:
    """Create a copy of the object.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
        name (str, optional): New name for the copied object. Defaults to None.
        collection (bpy.types.Collection, optional): Optional collection to put new object inside of. Defaults to None.
        is_library_object (bool, optional): Whether object is part of a linked library. Defaults to False.
        is_copy (bool, optional): Make a deep copy of the mesh data block
    Returns:
        bpy.types.Object: The newly created scene object.
    """
    obj = verify(obj)
    obj_data = obj.data
    if is_copy:
        obj_data = obj.data.copy()
    new_obj = bpy.data.objects.new(obj.name, obj_data)
    if name is not None:
        new_obj.name = name
    if collection is not None:
        collection.objects.link(new_obj)
    else:
        # Add to scene collection if no collection given
        scene = zpy.blender.verify_blender_scene()
        scene.collection.objects.link(obj)
    # TODO: Library Overriding functions
    if is_library_object:
        log.warning(f"Making mesh and material data local for obj {new_obj.name}")
        new_obj.data.make_local()
        for i in range(len(new_obj.material_slots)):
            bpy.data.objects[new_obj.name].material_slots[i].material.make_local()
        # Original object reference is lost if local copies are made
        new_obj = bpy.data.objects[new_obj.name]
    return new_obj


def translate(
    obj: Union[bpy.types.Object, str],
    translation: Union[Tuple[float], mathutils.Vector] = (0, 0, 0),
    is_absolute: bool = False,
) -> None:
    """Translate an object (in blender units).

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
        translation (Union[Tuple[float], mathutils.Vector], optional): Translation vector (x, y, z). Default (0, 0, 0).
        is_absolute: (bool, optional): The translation vector becomes the absolute world position

    """
    obj = verify(obj)
    log.debug(f"Translating object {obj.name} by {translation}")
    log.debug(f"Before - obj.matrix_world\n{obj.matrix_world}")
    if not isinstance(translation, mathutils.Vector):
        translation = mathutils.Vector(translation)
    if is_absolute:
        obj.location = translation
    else:
        obj.location = obj.location + translation
    log.debug(f"After - obj.matrix_world\n{obj.matrix_world}")


def rotate(
    obj: Union[bpy.types.Object, str],
    rotation: Union[Tuple[float], mathutils.Euler] = (0.0, 0.0, 0.0),
    axis_order: str = "XYZ",
) -> None:
    """Rotate the given object with Euler angles.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
        rotation (Union[Tuple[float], mathutils.Euler]): Rotation values in radians
        axis_order (str, optional): Axis order of rotation
    """
    obj = verify(obj)
    view_layer = zpy.blender.verify_view_layer()
    select(obj)
    log.info(f"Rotating object {obj.name} by {rotation} radians in {axis_order}. ")
    log.debug(f"Before - obj.matrix_world\n{obj.matrix_world}")
    if not isinstance(rotation, mathutils.Euler):
        rotation = mathutils.Euler(rotation)
    new_rotation_mat = rotation.to_matrix() @ obj.rotation_euler.to_matrix()
    new_rotation = new_rotation_mat.to_euler(axis_order)
    obj.rotation_euler = mathutils.Euler(new_rotation, axis_order)
    view_layer.update()
    log.debug(f"After - obj.matrix_world\n{obj.matrix_world}")


def scale(
    obj: Union[bpy.types.Object, str], scale: Tuple[float] = (1.0, 1.0, 1.0)
) -> None:
    """Scale an object.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
        scale (Tuple[float], optional): Scale for each axis (x, y, z). Defaults to (1.0, 1.0, 1.0).
    """
    obj = verify(obj)
    view_layer = zpy.blender.verify_view_layer()
    select(obj)
    log.info(f"Scaling object {obj.name} by {scale}")
    log.debug(f"Before - obj.matrix_world\n{obj.matrix_world}")
    bpy.ops.transform.resize(value=scale)
    view_layer.update()
    log.debug(f"After - obj.matrix_world\n{obj.matrix_world}")


def jitter_mesh(
    obj: Union[bpy.types.Object, str],
    scale: Tuple[float] = (0.01, 0.01, 0.01),
) -> None:
    """Randomize the vertex coordinates of a mesh object.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
        scale (Tuple[float], optional): Scale for vertex offset in each axis (x, y, z). Defaults to (0.01, 0.01, 0.01).
    """
    obj = verify(obj)
    if not obj.type == "MESH":
        log.warning("Jitter mesh requires object to be of type MESH")
        return
    for vertex in obj.data.vertices:
        offset = mathutils.Vector(
            (
                random.uniform(-1.0, 1.0) * obj.dimensions.x * scale[0],
                random.uniform(-1.0, 1.0) * obj.dimensions.y * scale[1],
                random.uniform(-1.0, 1.0) * obj.dimensions.z * scale[2],
            )
        )
        vertex.co += offset


def jitter(
    obj: Union[bpy.types.Object, str],
    translate_range: Tuple[Tuple[float]] = (
        (0, 0),
        (0, 0),
        (0, 0),
    ),
    rotate_range: Tuple[Tuple[float]] = (
        (0, 0),
        (0, 0),
        (0, 0),
    ),
    scale_range: Tuple[Tuple[float]] = (
        (1.0, 1.0),
        (1.0, 1.0),
        (1.0, 1.0),
    ),
) -> None:
    """Apply random scale (blender units) and rotation (radians) to object.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
        translate_range (Tuple[Tuple[float]], optional): (min, max) of uniform noise on translation in (x, y, z) axes.
            Defaults to ( (-0.05, 0.05), (-0.05, 0.05), (-0.05, 0.05), ).
        rotate_range (Tuple[Tuple[float]], optional): (min, max) of uniform noise on rotation in (x, y, z) axes.
            Defaults to ( (-0.05, 0.05), (-0.05, 0.05), (-0.05, 0.05), ).
        scale_range (Tuple[Tuple[float]], optional): (min, max) of uniform noise on scale in (x, y, z) axes.
            Defaults to ( (1.0, 1.0), (1.0, 1.0), (1.0, 1.0), ).
    """
    obj = verify(obj)
    translate(
        obj,
        translation=(
            random.uniform(translate_range[0][0], translate_range[0][1]),
            random.uniform(translate_range[1][0], translate_range[1][1]),
            random.uniform(translate_range[2][0], translate_range[2][1]),
        ),
    )
    rotate(
        obj,
        rotation=(
            random.uniform(rotate_range[0][0], rotate_range[0][1]),
            random.uniform(rotate_range[1][0], rotate_range[1][1]),
            random.uniform(rotate_range[2][0], rotate_range[2][1]),
        ),
    )
    scale(
        obj,
        scale=(
            random.uniform(scale_range[0][0], scale_range[0][1]),
            random.uniform(scale_range[1][0], scale_range[1][1]),
            random.uniform(scale_range[2][0], scale_range[2][1]),
        ),
    )


_SAVED_POSES = {}


def save_pose(
    obj: Union[bpy.types.Object, str],
    pose_name: str = None,
) -> None:
    """Save a pose (rot and pos) to dict.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
        pose_name (str): Name of saved pose (will be stored in internal SAVED_POSES dict)
    """
    obj = verify(obj)
    log.info(f"Saving pose {pose_name} based on object {obj.name}")
    if pose_name is None:
        pose_name = obj.name
    _SAVED_POSES[pose_name] = obj.matrix_world.copy()


def restore_pose(
    obj: Union[bpy.types.Object, str],
    pose_name: str = None,
) -> None:
    """Restore an object to a position.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)
        pose_name (str): Name of saved pose (must be in internal SAVED_POSES dict)
    """
    obj = verify(obj)
    log.info(f"Restoring pose {pose_name} to object {obj.name}")
    if pose_name is None:
        pose_name = obj.name
    obj.matrix_world = _SAVED_POSES[pose_name]


def lighting_randomize(
    energy_jitter: bool = True,
    jitter: bool = True,
    energy_range_sun: Tuple[int] = (1, 20),
    energy_range_point: Tuple[int] = (100, 500),
) -> None:
    """Randomizes Lighting.
    Args:
        energy_jitter: bool: Whether to jitter the lighting intensity for lights in scene. Defaults to True
        energy_range_sun: Tuple[int]: Range for intensity for SUN lights in scene.
        energy_range_point: Tuple[int]: Range for intensity for POINT, SPOT, and AREA lights in scene.
    """
    # check if lights are in the scene, if not log error
    for obj in bpy.context.scene.objects:
        if obj.type == "LIGHT":
            continue
        else:
            log.debug("add lights to use this function")
    # Loop through objects in scene and randomly toggle them on and off in the render,
    # these will still be visible in preview scene
    for obj in bpy.data.lights:
        if obj.type == "POINT" or obj.type == "SPOT" or obj.type == "AREA":
            bpy.data.objects[obj.name].hide_render = bool(random.randint(0, 1))
            if energy_jitter:
                bpy.data.objects[obj.name].data.energy = random.randint(
                    *energy_range_point
                )
        if obj.type == "SUN":
            bpy.data.objects[obj.name].hide_render = bool(random.randint(0, 1))
            if energy_jitter:
                bpy.data.objects[obj.name].data.energy = random.randint(
                    *energy_range_sun
                )
            # bpy.data.scenes["Scene"].world.use_nodes = True
        if jitter:
            zpy.objects.jitter(
                obj.name,
                translate_range=(
                    (-2, 2),
                    (-2, -2),
                    (1, 5),
                ),
            )



from zpy.files import to_pathlib_path
import os
import yaml

ENDPOINT = "https://ragnarok.zumok8s.org"
CONFIG_FILE = "~/.zpy/config.yaml"


def initialize_config():
    """initialize config

    If CONFIG_FILE doesnt exist write it and put in prod as the endpoint. Also creates
    the ~/.zpy folder if not existing. The config is some variables needed by the cli to
    make validated requests to the backend.
    """
    path = to_pathlib_path(os.path.expanduser(CONFIG_FILE))
    if path.exists():
        return
    CONFIG = {"ENVIRONMENT": "prod", "TOKEN": None, "ENDPOINT": ENDPOINT}
    path.parent.mkdir(parents=True, exist_ok=True)
    write_config(CONFIG)


def read_config(file=CONFIG_FILE):
    """read config

    Read zpy cli configuration file.

    Args:
        env: which enviroment to read config for
    Returns:
        config: dictionary of current configuration
    """
    path = to_pathlib_path(os.path.expanduser(file))
    with path.open() as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    return config


def write_config(config, file=CONFIG_FILE):
    """write config

    Write zpy cli configuration file.

    Args:
        config (dict): new configuration to write
    """
    path = to_pathlib_path(os.path.expanduser(file))
    with path.open("w") as f:
        yaml.dump(config, f)


def add_env(name, endpoint):
    """add environment

    Add a new environment configuration file.

    Args:
        name: name of the environment
        endpoint: endpoint for the new enviroment
    """
    new_config = {"ENVIRONMENT": name, "TOKEN": None, "ENDPOINT": endpoint}
    write_config(new_config, file=f"~/.zpy/{name}.yaml")


def swap_env(name):
    """swap environment

    Swap the current environment configuration.

    Args:
        name: swap to this env
    """
    old_config = read_config()
    new_config = read_config(file=f"~/.zpy/{name}.yaml")
    write_config(new_config)
    write_config(old_config, file=f"~/.zpy/{old_config['ENVIRONMENT']}.yaml")



"""
    Image utilties.
"""
import logging
from itertools import groupby
from pathlib import Path
from typing import Dict, List, Union

import numpy as np
from PIL import Image
from scipy import ndimage as ndi
from shapely.geometry import MultiPolygon, Polygon
from skimage import color, exposure, img_as_uint, io, measure
from skimage.morphology import binary_closing, binary_opening
from skimage.transform import resize

import gin
import zpy

log = logging.getLogger(__name__)


def open_image(
    image_path: Union[Path, str],
) -> np.ndarray:
    """Open image from path to ndarray.

    Args:
        image_path (Union[Path, str]): Path to image.

    Returns:
        np.ndarray: Image as numpy array.
    """
    image_path = zpy.files.verify_path(image_path, make=False)
    img = None
    try:
        img = io.imread(image_path)
        if img.shape[2] > 3:
            log.debug("RGBA image detected!")
            img = img[:, :, :3]
        if img.max() > 2.0:
            img = np.divide(img, 255.0)
    except Exception as e:
        log.error(f"Error {e} when opening {image_path}")
    return img


def remove_alpha_channel(image_path: Union[Path, str]) -> None:
    """Remove the alpha channel in an image (overwrites image).

    Args:
        image_path (Union[Path, str]): Path to image.
    """
    img = open_image(image_path)
    io.imsave(image_path, img)
    log.info(f"Saving image with no alpha channel at {image_path}")


@gin.configurable
def jpeg_compression(
    image_path: Union[Path, str],
    quality: int = 40,
) -> Path:
    """Add jpeg compression to an image (overwrites image).

    Args:
        image_path (Union[Path, str]): Path to image.
        quality (int, optional): Compression quality. Defaults to 40.

    Returns:
        Path: Path to image.
    """
    image_path = zpy.files.verify_path(image_path, make=False)
    img = io.imread(image_path)
    # img = Image.open(image_path)
    # Make sure image is jpeg
    if not image_path.suffix == ".jpeg":
        image_path = image_path.with_suffix(".jpeg")
    io.imsave(image_path, arr=img, quality=quality)
    # img.save(image_path, "JPEG")
    log.info(f"Saving compressed image at {image_path}")
    return image_path


@gin.configurable
def resize_image(
    image_path: Union[Path, str],
    width: int = 640,
    height: int = 480,
) -> Path:
    """Resize an image (overwrites image).

    Args:
        image_path (Union[Path, str]): Path to image.
        width (int, optional): Width of image. Defaults to 640.
        height (int, optional): Height of image. Defaults to 480.

    Returns:
        Path: Path to image.
    """
    img = open_image(image_path)
    resized_img = resize(img, (height, width), anti_aliasing=True)
    io.imsave(image_path, resized_img)


def pixel_mean_std(
    flat_images: List[np.ndarray],
) -> Dict:
    """Return the pixel mean and std from a flattened images array.

    Args:
        flat_images (List[np.ndarray]): Image pixels in a flattened array

    Returns:
        Dict: Pixel means and std as floats and integers (256)
    """
    # HACK: Incorrect type assumption
    flat_images = flat_images[0]
    if np.amax(flat_images) > 1:
        std_256 = np.std(flat_images, axis=0)
        mean_256 = np.mean(flat_images, axis=0)
        std = std_256 / 256
        mean = mean_256 / 256
    else:
        std = np.std(flat_images, axis=0)
        mean = np.mean(flat_images, axis=0)
        std_256 = std * 256.0
        mean_256 = mean * 256.0
    return {
        "mean": mean,
        "std": std,
        "mean_256": mean_256,
        "std_256": std_256,
    }


def flatten_images(
    images: List[np.ndarray],
    max_pixels: int = 500000,
) -> List[np.ndarray]:
    """Flatten a list of images in ndarray form.

    Args:
        images (List[np.ndarray]): List of images in ndarray form.
        max_pixels (int, optional): Maximum number of pixels in the flattened array. Defaults to 500000.

    Returns:
        List[np.ndarray]: List of flattened images.
    """
    flat_images = []
    for image in images:
        dims = np.shape(image)
        if len(dims) == 3:
            flat_images.append(np.reshape(image, (dims[0] * dims[1], dims[2])))
    flat_images = np.concatenate(flat_images, axis=0)
    subsample = flat_images[np.random.randint(flat_images.shape[0], size=max_pixels), :]
    return [subsample]


def pad_with(vector, pad_width, iaxis, kwargs):
    """Pad a vector.

    https://numpy.org/doc/stable/reference/generated/numpy.pad.html
    """
    pad_value = kwargs.get("padder", 10)
    vector[: pad_width[0]] = pad_value
    vector[-pad_width[1] :] = pad_value


def binary_mask_to_rle(binary_mask) -> Dict:
    """Converts a binary mask to a RLE (run-length-encoded) dictionary.

    https://stackoverflow.com/questions/49494337/encode-numpy-array-using-uncompressed-rle-for-coco-dataset
    """
    binary_mask = np.asfortranarray(binary_mask)
    rle = {"counts": [], "size": list(binary_mask.shape)}
    counts = rle.get("counts")
    for i, (value, elements) in enumerate(groupby(binary_mask.ravel(order="F"))):
        if i == 0 and value == 1:
            counts.append(0)
        counts.append(len(list(elements)))
    return rle


@gin.configurable
def seg_to_annotations(
    image_path: Union[Path, str],
    remove_salt: bool = True,
    rle_segmentations: bool = False,
    float_annotations: bool = False,
    max_categories: int = 1000,
) -> List[Dict]:
    """Convert a segmentation image into bounding boxes and polygon segmentations.

    Args:
        image_path (Union[Path, str]): Path to image.
        remove_salt (bool, optional): Remove salt when calculating bounding box and polygons. Defaults to True.
        rle_segmentations (bool, optional): Include RLE polygons in annotation dictionaries. Defaults to False.
        float_annotations (bool, optional): Include float (0, 1) bboxes/polygons in annotation dicts. Defaults False.
        max_categories (int, optional): Maximum number of categories allowed in an image. Defaults to 1000.

    Raises:
        ValueError: Too many categories (usually means segmentation image is not single colors)

    Returns:
        List[Dict]: List of annotation dictionaries.
    """
    log.info(f"Extracting annotations from segmentation: {image_path}")
    image_path = zpy.files.verify_path(image_path, make=False)
    img = open_image(image_path)
    img_height, img_width = img.shape[0], img.shape[1]
    # Unique colors represent each unique category
    unique_colors = np.unique(img.reshape(-1, img.shape[2]), axis=0)
    # Store bboxes, seg polygons, and area in annotations list
    annotations = []
    # Loop through each category
    if unique_colors.shape[0] > max_categories:
        raise ValueError(f"Over {max_categories} categories: {unique_colors.shape[0]}")
    for i in range(unique_colors.shape[0]):
        seg_color = unique_colors[i, :]
        log.debug(f"Unique color {seg_color}")
        if all(np.equal(seg_color, np.zeros(3))):
            log.debug("Color is background.")
            continue
        # Make an image mask for this category
        masked_image = img.copy()
        mask = (img != seg_color).any(axis=-1)
        masked_image[mask] = np.zeros(3)
        masked_image = color.rgb2gray(masked_image)
        if log.getEffectiveLevel() == logging.DEBUG:
            masked_image_name = (
                str(image_path.stem) + f"_masked_{i}" + str(image_path.suffix)
            )
            masked_image_path = image_path.parent / masked_image_name
            io.imsave(
                masked_image_path, img_as_uint(exposure.rescale_intensity(masked_image))
            )
        if remove_salt:
            # Remove "salt"
            # https://scikit-image.org/docs/dev/api/skimage.morphology
            masked_image = binary_opening(masked_image)
            masked_image = binary_closing(masked_image)
        else:
            masked_image = binary_opening(masked_image)
        # HACK: Pad masked image so segmented objects that extend beyond
        #       image are properly contoured
        masked_image = np.pad(masked_image, 1, pad_with, padder=False)
        # RLE encoded segmentation from binary image
        if rle_segmentations:
            rle_segmentation = binary_mask_to_rle(masked_image)
        # Fill in the holes
        filled_masked_image = ndi.binary_fill_holes(masked_image)
        # Get countours for each blob
        contours = measure.find_contours(
            filled_masked_image, 0.01, positive_orientation="low"
        )
        log.debug(f"found {len(contours)} contours for {seg_color} in {image_path}")
        # HACK: Sometimes all you get is salt for an image, in this case
        #       do not add any annotation for this category.
        if len(contours) == 0:
            continue
        segmentations = []
        segmentations_float = []
        bboxes = []
        bboxes_float = []
        areas = []
        areas_float = []
        polygons = []
        for contour in contours:
            # Flip from (row, col) representation to (x, y)
            # and subtract the padding pixel
            for j in range(len(contour)):
                row, col = contour[j]
                contour[j] = (col - 1, row - 1)
            # Make a polygon and simplify it
            poly = Polygon(contour)
            poly = poly.simplify(1.0, preserve_topology=True)
            polygons.append(poly)
            # Segmentation
            segmentation = np.array(poly.exterior.coords).ravel().tolist()
            segmentations.append(segmentation)
            segmentations_float.append(
                [
                    x / img_height if k % 2 == 0 else x / img_width
                    for k, x in enumerate(segmentation)
                ]
            )
            # Bounding boxes
            x, y, max_x, max_y = poly.bounds
            bbox = (x, y, max_x - x, max_y - y)
            bbox_float = [
                bbox[0] / img_width,
                bbox[1] / img_height,
                bbox[2] / img_width,
                bbox[3] / img_height,
            ]
            bboxes.append(bbox)
            bboxes_float.append(bbox_float)
            # Areas
            areas.append(poly.area)
            areas_float.append(poly.area / (img_width * img_height))
        # Combine the polygons to calculate the bounding box and area
        multi_poly = MultiPolygon(polygons)
        x, y, max_x, max_y = multi_poly.bounds
        bbox = (x, y, max_x - x, max_y - y)
        area = multi_poly.area
        bbox_float = [
            bbox[0] / img_width,
            bbox[1] / img_height,
            bbox[2] / img_width,
            bbox[3] / img_height,
        ]
        area_float = area / (img_width * img_height)
        annotation = {
            "color": tuple(seg_color),
            # COCO standards
            "segmentation": segmentations,
            "bbox": bbox,
            "area": area,
            # List of list versions
            "bboxes": bboxes,
            "areas": areas,
        }
        if rle_segmentations:
            annotation["segmentation_rle"] = rle_segmentation
        if float_annotations:
            annotation["segmentation_float"] = segmentations_float
            annotation["bbox_float"] = bbox_float
            annotation["area_float"] = area_float
            annotation["bboxes_float"] = bboxes_float
            annotation["areas_float"] = areas_float
        annotations.append(annotation)
    return annotations



"""
    Script loading/running panel and functions.
"""
import importlib
import logging
from pathlib import Path

import bpy
import zpy

log = logging.getLogger(__name__)

if "bpy" in locals():
    importlib.reload(zpy)


def registerSceneProperties():
    """Properties applied to scenes."""
    bpy.types.Scene.zpy_gin_config_path = bpy.props.StringProperty(
        name="",
        description="Path to a gin config file.",
        default="",
        subtype="FILE_PATH",
        update=_load_gin_config,
    )
    bpy.types.Scene.zpy_runpy_path = bpy.props.StringProperty(
        name="",
        description="Path to run.py file",
        default="",
        subtype="FILE_PATH",
        update=_load_runpy,
    )
    bpy.types.Scene.zpy_template_dir = bpy.props.StringProperty(
        name="",
        description="Path to script template directory.",
        default=str(zpy.assets.script_template_dir()),
        subtype="DIR_PATH",
    )


def _load_gin_config(self, context) -> None:
    """Load gin config from file."""
    bpy.ops.scene.zpy_load_gin_config()


def _load_runpy(self, context) -> None:
    """Load run.py from file."""
    bpy.ops.scene.zpy_load_runpy()


class LoadGinConfigOperator(bpy.types.Operator):
    """Load gin config from file."""

    bl_idname = "scene.zpy_load_gin_config"
    bl_label = "Load gin config from file."
    bl_description = "Load gin config from file."
    bl_category = "ZPY"
    bl_options = {"REGISTER"}

    # Default name of the texts in Blender when loading
    DEFAULT_TEXT_NAME = "config"

    def execute(self, context):
        zpy.blender.load_text_from_file(
            bpy.path.abspath(context.scene.zpy_gin_config_path),
            text_name=self.DEFAULT_TEXT_NAME,
        )
        return {"FINISHED"}


class PushGinConfigOperator(bpy.types.Operator):
    """Push gin config to file."""

    bl_idname = "scene.zpy_push_gin_config"
    bl_label = "Push gin config to file."
    bl_description = "Push gin config to file."
    bl_category = "ZPY"
    bl_options = {"REGISTER"}

    def execute(self, context):
        _text = bpy.data.texts[LoadGinConfigOperator.DEFAULT_TEXT_NAME].as_string()
        with open(bpy.path.abspath(context.scene.zpy_gin_config_path), "w") as _file:
            _file.write(_text)
        return {"FINISHED"}


class LoadRunpyOperator(bpy.types.Operator):
    """Load run.py from file."""

    bl_idname = "scene.zpy_load_runpy"
    bl_label = "Load run.py from file."
    bl_description = "Load run.py from file."
    bl_category = "ZPY"
    bl_options = {"REGISTER"}

    # Default name of the texts in Blender when loading
    DEFAULT_TEXT_NAME = "run"

    def execute(self, context):
        zpy.blender.load_text_from_file(
            bpy.path.abspath(context.scene.zpy_runpy_path),
            text_name=self.DEFAULT_TEXT_NAME,
        )
        return {"FINISHED"}


class PushRunpyOperator(bpy.types.Operator):
    """Push run.py to file."""

    bl_idname = "scene.zpy_push_runpy"
    bl_label = "Push run.py to file."
    bl_description = "Push run.py to file."
    bl_category = "ZPY"
    bl_options = {"REGISTER"}

    def execute(self, context):
        _text = bpy.data.texts[LoadRunpyOperator.DEFAULT_TEXT_NAME].as_string()
        with open(bpy.path.abspath(context.scene.zpy_runpy_path), "w") as _file:
            _file.write(_text)
        return {"FINISHED"}


class SCENE_PT_ScriptPanel(bpy.types.Panel):
    """UI for the addon that is visible in Blender."""

    bl_idname = "SCENE_PT_ScriptPanel"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Script"
    bl_category = "ZPY"
    bl_options = {"DEFAULT_CLOSED"}

    def draw(self, context):
        layout = self.layout
        scene = context.scene

        row = layout.row()
        row.operator(
            "scene.zpy_load_templates",
            text="Load Templates",
            icon="TEXT",
        )
        row = layout.row()
        row.label(text="Run.py Path")
        row = layout.row()
        row.prop(scene, "zpy_runpy_path", expand=True)
        row = layout.row()
        row.operator(
            "scene.zpy_load_runpy",
            text="Pull",
            icon="IMPORT",
        )
        row.operator(
            "scene.zpy_push_runpy",
            text="Push",
            icon="EXPORT",
        )
        row = layout.row()
        row.label(text="Gin Config Path")
        row = layout.row()
        row.prop(scene, "zpy_gin_config_path")
        row = layout.row()
        row.operator(
            "scene.zpy_load_gin_config",
            text="Pull",
            icon="IMPORT",
        )
        row.operator(
            "scene.zpy_push_gin_config",
            text="Push",
            icon="EXPORT",
        )


class LoadTemplatesOperator(bpy.types.Operator):
    """Loads templates for run.py and gin config."""

    bl_idname = "scene.zpy_load_templates"
    bl_label = "Loads templates for run.py and gin config."
    bl_description = "Loads templates for run.py and gin config."
    bl_category = "ZPY"
    bl_options = {"REGISTER"}

    def execute(self, context):
        bpy.ops.text.zpy_load_ginconfig_template()
        bpy.ops.text.zpy_load_runpy_template()
        return {"FINISHED"}


class TEXT_PT_LoadRunPyTemplateOperator(bpy.types.Operator):
    """Opens the run.py template."""

    bl_idname = "text.zpy_load_runpy_template"
    bl_label = "Open zpy run.py template."

    def execute(self, context):
        template_dir = Path(bpy.path.abspath(context.scene.zpy_template_dir))
        template_path = template_dir / "run.py"
        zpy.blender.load_text_from_file(
            template_path,
            text_name=LoadRunpyOperator.DEFAULT_TEXT_NAME,
            open_text=True,
        )
        return {"FINISHED"}


class TEXT_PT_LoadGinConfigTemplateOperator(bpy.types.Operator):
    """Opens the gin config template."""

    bl_idname = "text.zpy_load_ginconfig_template"
    bl_label = "Open gin config template."

    def execute(self, context):
        template_dir = Path(bpy.path.abspath(context.scene.zpy_template_dir))
        template_path = template_dir / "config.gin"
        zpy.blender.load_text_from_file(
            template_path,
            text_name=LoadGinConfigOperator.DEFAULT_TEXT_NAME,
            open_text=True,
        )
        return {"FINISHED"}


def script_template_menu(self, context):
    self.layout.separator()
    self.layout.operator("text.zpy_load_runpy_template", text="(zpy) Run Script")
    self.layout.operator("text.zpy_load_ginconfig_template", text="(zpy) Config Text")



import json

import requests

from cli.utils import fetch_auth


@fetch_auth
def fetch_projects(filters, url, auth_headers):
    """fetch projects

    Fetch projects from ZumoLabs backend.

    Args:
        filters (dict): query param filters for API call
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    endpoint = f"{url}/api/v1/projects/"
    r = requests.get(endpoint, headers=auth_headers, params=filters)
    if r.status_code != 200:
        r.raise_for_status()
    return json.loads(r.text)["results"]


@fetch_auth
def create_project(account_id, name, url, auth_headers):
    """
    create project

    Create empty project on ZumoLabs backend.

    Args:
        account_id (str): uuid of account
        name (str): name of project
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    endpoint = f"{url}/api/v1/projects/"
    r = requests.post(
        endpoint,
        data={"account": account_id, "name": name},
        headers=auth_headers,
    )
    if r.status_code != 201:
        r.raise_for_status()



"""
    Image version of Saver object.
"""
import logging
from pathlib import Path
from typing import Dict, Tuple, Union

import gin
import numpy as np

import zpy

log = logging.getLogger(__name__)


@gin.configurable
class ImageSaver(zpy.saver.Saver):
    """Saver class for Image based datasets."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.images = {}
        self.image_name_to_id = {}
        self.seg_annotations_color_to_id = {}

    @gin.configurable
    def add_image(
        self,
        name: str = "default image",
        style: str = "default",
        output_path: Union[Path, str] = "/tmp/test.png",
        frame: int = 0,
        width: int = 640,
        height: int = 480,
        zero_indexed: bool = True,
        **kwargs,
    ) -> Dict:
        """Add a new image annotation to the Saver object.

        Pass any additional keys you want in the image annotation dict as kwargs.

        Args:
            name (str, optional): Unique image name. Defaults to 'default image'.
            style (str, optional): Type of image in [default, segmenation, depth]. Defaults to 'default'.
            output_path (Union[Path, str], optional): Path to image file. Defaults to '/tmp/test.png'.
            frame (int, optional): Frame is used to link images taken at the same moment in time. Defaults to 0.
            width (int, optional): Width of image. Defaults to 640.
            height (int, optional): Height of image. Defaults to 480.
            zero_indexed (bool, optional): Whether image id is zero indexed. Defaults to True.

        Returns:
            Dict: The image annotation dictionary.
        """
        image = {
            "name": name,
            "style": style,
            "output_path": str(output_path),
            "relative_path": str(Path(output_path).relative_to(self.output_dir)),
            "frame": frame,
            "width": width,
            "height": height,
        }
        image.update(**kwargs)
        image["id"] = len(self.images.keys())
        image["id"] += 0 if zero_indexed else 1
        log.debug(f"Adding image: {zpy.files.pretty_print(image)}")
        self.images[image["id"]] = image
        self.image_name_to_id[name] = image["id"]
        return image

    @gin.configurable
    def add_annotation(
        self,
        image: str = "default image",
        category: str = None,
        subcategory: str = None,
        subcategory_zero_indexed: bool = True,
        seg_image: str = None,
        seg_color: Tuple[float] = None,
        parse_on_add: bool = True,
        **kwargs,
    ) -> Dict:
        """Add a new annotation to the Saver object.

        Pass any additional keys you want in the annotation dict as kwargs.

        Args:
            image (str, optional): Unique image name. Defaults to 'default image'.
            category (str, optional): Name of category. Defaults to None.
            subcategory (str, optional): Name of subcategory. Defaults to None.
            subcategory_zero_indexed (bool, optional): Whether subcategories are zero-indexed. Defaults to True.
            seg_image (str, optional): Name of segmentation image that corresponds to this image. Defaults to None.
            seg_color (Tuple[float], optional): Segmentation color of entity in this annotation. Defaults to None.
            parse_on_add (bool, optional): Calculate bounding box and polygons and populate segmenation fields.
                Defaults to True.

        Returns:
            Dict: The annotation dictionary.
        """
        image_id = self.image_name_to_id.get(image, None)
        assert image_id is not None, f"Could not find id for image {image}"
        assert category is not None, "Must provide a category for annotation."
        category_id = self.category_name_to_id.get(category, None)
        assert category_id is not None, f"Could not find id for category {category}"
        self.categories[category_id]["count"] += 1
        annotation = {
            "image_id": image_id,
            "category_id": category_id,
        }
        if subcategory is not None:
            subcategory_id = self.categories[category_id]["subcategories"].index(
                subcategory
            )
            self.categories[category_id]["subcategory_count"][subcategory_id] += 1
            subcategory_id += 0 if subcategory_zero_indexed else 1
            annotation["subcategory_id"] = subcategory_id
        annotation.update(**kwargs)
        annotation["id"] = len(self.annotations)
        log.info(f"Adding annotation: {zpy.files.pretty_print(annotation)}")
        # For segmentation images, add bbox/poly/mask annotation
        if seg_image is not None and seg_color is not None:
            seg_image_id = self.image_name_to_id.get(seg_image, None)
            assert seg_image_id is not None, f"Could not find id for image {seg_image}"
            annotation["seg_color"] = seg_color
            if self.seg_annotations_color_to_id.get(seg_image, None) is None:
                self.seg_annotations_color_to_id[seg_image] = {}
            self.seg_annotations_color_to_id[seg_image][seg_color] = annotation["id"]
        self.annotations.append(annotation)
        # This call creates correspondences between segmentation images
        # and the annotations. It should be used after both the images
        # and annotations have been added to the saver.
        if parse_on_add:
            self.parse_annotations_from_seg_image(image_name=seg_image)
        return annotation

    def parse_annotations_from_seg_image(
        self,
        image_name: str,
    ) -> None:
        """Populate annotation field based on segmentation image.

        Args:
            image_name (str): Name of image in which to put parse out segmentations.

        Raises:
            ValueError: Image is not a segmentation image.
        """
        # Verify that file is segmentation image
        is_iseg = zpy.files.file_is_of_type(image_name, "instance segmentation image")
        is_cseg = zpy.files.file_is_of_type(image_name, "class segmentation image")
        if not (is_iseg or is_cseg):
            raise ValueError("Image is not segmentation image")
        seg_image_id = self.image_name_to_id.get(image_name, None)
        assert seg_image_id is not None, f"Could not find id for image {image_name}"
        image_path = self.images[seg_image_id]["output_path"]
        if self.seg_annotations_color_to_id.get(image_name, None) is None:
            log.warning(f"No annotations found for {image_name}")
        for annotation in zpy.image.seg_to_annotations(image_path):
            if (
                self.seg_annotations_color_to_id[image_name].get(
                    annotation["color"], None
                )
                is None
            ):
                log.warning(
                    f'No annotations found for color {annotation["color"]} in {image_name}'
                )
                log.warning(
                    f"Available colors are {list(self.seg_annotations_color_to_id[image_name].keys())}"
                )
                closest_color = zpy.color.closest_color(
                    annotation["color"],
                    list(self.seg_annotations_color_to_id[image_name].keys()),
                )
                if closest_color is None:
                    log.warning("Could not find close enough color, skipping ...")
                    continue
                else:
                    log.warning(f"Using closest color {closest_color}")
                    idx = self.seg_annotations_color_to_id[image_name][closest_color]
            else:
                idx = self.seg_annotations_color_to_id[image_name][annotation["color"]]
            self.annotations[idx].update(annotation)

    @gin.configurable
    def output_annotated_images(
        self,
        num_annotated_images: int = 10,
    ) -> None:
        """Dump annotated sampled images to the meta folder.

        Args:
            num_annotated_images (int, optional): Number of annotation images to output. Defaults to 10.
        """
        log.info("Output annotated images...")
        import zpy.viz

        output_path = self.output_dir / self.HIDDEN_METAFOLDER_FILENAME
        output_path = zpy.files.verify_path(output_path, make=True, check_dir=True)
        for i, image in enumerate(self.images.values()):
            # Annotation images take a while
            if i >= num_annotated_images:
                return
            annotations = []
            for annotation in self.annotations:
                if annotation["image_id"] == image["id"]:
                    annotations.append(annotation)
            if len(annotations) > 0:
                zpy.viz.draw_annotations(
                    image_path=Path(image["output_path"]),
                    annotations=annotations,
                    categories=self.categories,
                    output_path=output_path,
                )

    @gin.configurable
    def output_meta_analysis(
        self,
        image_sample_size: int = 50,
    ) -> None:
        """Perform a full meta analysis, outputting some meta files.

        Args:
            image_sample_size (int, optional): How many images to sample for meta analysis. Defaults to 50.
        """
        log.info(f"perform meta analysis image_sample_size:{image_sample_size}...")

        import zpy.files

        image_paths = [
            i["output_path"] for i in self.images.values() if i["style"] == "default"
        ]
        image_paths = zpy.files.sample(image_paths, sample_size=image_sample_size)
        opened_images = [zpy.image.open_image(i) for i in image_paths]
        flat_images = zpy.image.flatten_images(opened_images)
        pixel_mean_std = zpy.image.pixel_mean_std(flat_images)

        meta_dict = {
            "number_images": len(self.images),
            "number_annotations": len(self.annotations),
            "number_categories": len(self.categories),
            "category_names": [c["name"] for c in self.categories.values()],
            "pixel_mean": np.array2string(pixel_mean_std["mean"], precision=2),
            "pixel_std": np.array2string(pixel_mean_std["std"], precision=2),
            "pixel_256_mean": np.array2string(pixel_mean_std["mean_256"], precision=0),
            "pixel_256_std": np.array2string(pixel_mean_std["std_256"], precision=0),
        }

        output_path = self.output_dir / self.HIDDEN_METAFOLDER_FILENAME
        output_path = zpy.files.verify_path(output_path, make=True, check_dir=True)
        self.write_datasheet(output_path / self.HIDDEN_DATASHEET_FILENAME, meta_dict)

        try:
            import zpy.viz

            zpy.viz.image_grid_plot(images=opened_images, output_path=output_path)
            zpy.viz.image_shape_plot(images=opened_images, output_path=output_path)
            zpy.viz.color_correlations_plot(
                flat_images=flat_images, output_path=output_path
            )
            zpy.viz.pixel_histograms(flat_images=flat_images, output_path=output_path)
            zpy.viz.category_barplot(
                categories=self.categories, output_path=output_path
            )
        except Exception as e:
            log.warning(f"Error when visualizing {e}")
            pass
